import numpy as np
import pandas as pd

from ..utils import shape_checker, type_checker, value_checker
from . import _utils_klnmf
from .nmf import NMF

EPSILON = np.finfo(np.float32).eps


class KLNMF(NMF):
    """
    Decompose a mutation count matrix X into the product of a signature
    matrix W and an exposure matrix H by minimizing the generalized
    Kullback-Leibler (KL) divergence under the constraint of having
    normalized signatures.

    Parameters
    ----------
    n_signatures: int
        Number of signatures to decipher.

    init_method : str, default=nndsvd
        One of "custom", "flat", "hierarchical_cluster", "nndsvd",
        "nndsvda", "nndsvdar" "random" and "separableNMF". Please see the initialization
        module for further details on each method.

    update_method : str, default=mu-joint
        One of "mu-standard" and "mu-joint". The standard multiplicative
        update rules alternates between optimizing the signatures and exposures.
        The joint multiplicative update rule updates both matrices at once.
        It requires one matrix multiplication less and is faster.

    min_iterations : int, default=500
        Minimum number of iterations.

    max_iterations : int, default=10000
        Maximum number of iterations.

    conv_test_freq: int
        The frequency at which the algorithm is tested for convergence.
        The objective function value is only computed every 'conv_test_freq'
        many iterations, which also affects a potentially saved history of
        the objective function values.

    tol : float, default=1e-7
        Tolerance of the stopping condition.

    Reference
    ---------
    D. Lee, H. Seung: Algorithms for Non-negative Matrix Factorization
    - Advances in neural information processing systems, 2000
    https://proceedings.neurips.cc/paper_files/paper/2000/file/f9d1152547c0bde01830b7e8bd60024c-Paper.pdf
    """

    def __init__(
        self,
        n_signatures=1,
        init_method="nndsvd",
        update_method="mu-joint",
        min_iterations=500,
        max_iterations=10000,
        conv_test_freq=10,
        tol=1e-7,
    ):
        super().__init__(
            n_signatures,
            init_method,
            min_iterations,
            max_iterations,
            conv_test_freq,
            tol,
        )
        value_checker("update method", update_method, ["mu-standard", "mu-joint"])
        self.update_method = update_method
        self.weights_kl = None
        self.weights_l_half = None

    @property
    def reconstruction_error(self) -> float:
        """
        The unweighted Kullback-Leibler divergence.
        """
        return _utils_klnmf.kl_divergence(self.X, self.W, self.H)

    @property
    def samplewise_reconstruction_error(self) -> np.ndarray:
        """
        The unweighted samplewise Kullback-Leibler divergence.
        """
        return _utils_klnmf.samplewise_kl_divergence(self.X, self.W, self.H)

    def objective_function(self) -> float:
        """
        The sum of the (weighted) Kullback-Leibler divergence and the sparsity
        penalty.
        """
        of_value = _utils_klnmf.kl_divergence(self.X, self.W, self.H, self.weights_kl)

        if self.weights_l_half is not None:
            of_value += np.dot(self.weights_l_half, np.sum(np.sqrt(self.H), axis=0))

        return of_value

    @property
    def objective(self) -> str:
        return "minimize"

    def loglikelihood(self) -> float:
        return _utils_klnmf.poisson_llh(self.X, self.W, self.H)

    def _update_W(self):
        self.W = _utils_klnmf.update_W(
            self.X, self.W, self.H, self.weights_kl, self.n_given_signatures
        )

    def _update_H(self):
        self.H = _utils_klnmf.update_H(
            self.X, self.W, self.H, self.weights_kl, self.weights_l_half
        )

    def _update_WH(self):
        if self.update_method == "mu-standard":
            self._update_H()
            if self.n_given_signatures < self.n_signatures:
                self._update_W()
        else:
            self.W, self.H = _utils_klnmf.update_WH(
                self.X,
                self.W,
                self.H,
                self.weights_kl,
                self.weights_l_half,
                self.n_given_signatures,
            )

    def _check_weights(self, weights: np.ndarray, name: str = "weights"):
        """
        Check if the given sample-specific loss function or l-1/2 penalty weights
        are compatible with the input data.

        weights : np.ndarray of shape (n_samples,)
            Sample-specific KL-divergence or sparsity penalty loss weights

        name : str, default='weights'
            Name to be displayed in a potential error message
        """
        type_checker(name, weights, np.ndarray)
        shape_checker(name, weights, (self.n_samples,))

        if not all(weights >= 0):
            raise ValueError(
                "Only non-negative KL-divergence and sparsity penalty weights "
                "are allowed."
            )

    def _setup_weight_parameters(
        self, weights_kl: np.ndarray, weights_l_half: np.ndarray
    ):
        for weights, name in zip(
            [weights_kl, weights_l_half], ["weights_kl", "weights_l_half"]
        ):
            if type(weights) in [float, int]:
                weights *= np.ones(self.n_samples)

            if type(weights) is list:
                weights = np.array(weights)

            if weights is not None:
                self._check_weights(weights, name)

            setattr(self, name, weights)

    def fit(
        self,
        data: pd.DataFrame,
        given_signatures=None,
        weights_kl=None,
        weights_l_half=None,
        init_kwargs=None,
        history=False,
        verbose=0,
    ):
        """
        Minimize the generalized Kullback-Leibler divergence D_KL(X || WH) between
        the mutation count matrix X and product of the signature matrix W and
        exposure matrix H under the constraint of normalized signatures.

        Parameters
        ----------
        data : pd.DataFrame
            The mutation count data

        given_signatures : pd.DataFrame, default=None
            Known signatures that should be fixed by the algorithm.
            The number of known signatures can be less or equal to the
            number of signatures specified in the algorithm instance.

        weights_kl : np.ndarray, default=None
            Per sample KL-divergence loss weights of shape (n_samples,).

        weights_l_half : np.ndarray, default=None
            Per sample l_half penalty weights of shape (n_samples,).
            They can be used to induce sparse exposures.

        init_kwargs : dict, default=None
            Any further keyword arguments to be passed to the initialization method.
            This includes, for example, a possible 'seed' keyword argument
            for all stochastic methods.

        history : bool, default=False
            If True, the objective function value will be stored after every
            'conv_test_freq' many iterations.

        verbose : int, default=0
            verbosity level

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        self._setup_data_parameters(data)
        self._setup_weight_parameters(weights_kl, weights_l_half)
        self._initialize(given_signatures, init_kwargs)
        of_values = [self.objective_function()]
        n_iteration = 0
        converged = False

        while not converged:
            n_iteration += 1

            if verbose and n_iteration % 1000 == 0:
                print(f"iteration: {n_iteration}; objective: {of_values[-1]:.2f}")

            self._update_WH()

            if n_iteration % self.conv_test_freq == 0:
                prev_of_value = of_values[-1]
                of_values.append(self.objective_function())
                rel_change = (prev_of_value - of_values[-1]) / prev_of_value
                converged = rel_change < self.tol and n_iteration >= self.min_iterations

            converged |= n_iteration >= self.max_iterations

        if history:
            self.history["objective_function"] = of_values[1:]

        return self



from abc import abstractmethod

import numpy as np
import pandas as pd

from ..utils import match_signatures_pair
from .initialization import initialize
from .signature_nmf import SignatureNMF

EPSILON = np.finfo(np.float32).eps


class NMF(SignatureNMF):
    """
    The abstract class NMF unifies the structure of NMF algorithms
    with a single signature matrix W and exposure matrix H.
    Examples of these algorithms include the standard NMF algorithm
    (Lee and Seung, 1999), minimum volume NMF (mvNMF) or NMF variants
    with regularizations on the entries of W and H.
    All of these NMF algorithms have the same parameters. Therefore,
    many properties of interest such as the signature correlation martrix
    or the sample embeddings are computed in the same manner.

    Overview:

    Every child class has to implement the following attributes:

        - reconstruction_error: float
            The reconstruction error between the count matrix and
            the reconstructed count matrix.

        - samplewise_reconstruction_error: np.ndarray
            The samplewise reconstruction error between the sample counts
            and the reconstructed sample counts.

        - objective: str
            "minimize" or "maximize". Whether the NMF algorithm maximizes or
            minimizes the objective function.
            Some algorithms maximize a likelihood, others minimize a distance.
            The distinction is useful for filtering NMF runs based on
            the fitted objective function value downstream.


    Every child class has to implement the following methods:

        - objective_function:
            The algorithm-specific objective function

        - loglikelihood:
            The loglikelihood of the underyling generative model

        - _update_W:
            update the signature matrix W

        - _update_H:
            update the exposure matrix H

        - fit:
            Apply the NMF algorithm for a given mutation count data or
            for given signatures and mutation count data


    The following attributes are implemented in the abstract class NMF:

        - signatures: pd.DataFrame
            The signature matrix including mutation type names and signature names

        - exposures: pd.DataFrame
            The exposure matrix including the signature names and sample names

        - _n_parameters:
            The number of parameters fitted

        - corr_signatures: pd.DataFrame
            The signature correlation matrix induced by their sample exposures

        - corr_samples: pd.DataFrame
            The sample correlation matrix induced by their signature exposures


    The following methods are implemented in the abstract class NMF:

        - initialize:
            Initialize all model parameters and latent variables depending on the
            initialization method chosen

        - _get_embedding_annotations:
            A helper function to get the sample names for the embedding plots

        - plot_embeddings:
            Plot signature or sample embeddings in 2D using PCA, tSNE or UMAP.
            The respective plotting functions are implemented in the plot.py module.

    More details are explained in the respective attributes and methods.
    """

    def __init__(
        self,
        n_signatures=1,
        init_method="nndsvd",
        min_iterations=500,
        max_iterations=10000,
        conv_test_freq=10,
        tol=1e-7,
    ):
        """
        Input:
        ------
        n_signatures: int
            The number of underlying signatures that are assumed to
            have generated the mutation count data.

        init_method: str
            One of "custom", "flat", "hierarchical_cluster", "nndsvd",
            "nndsvda", "nndsvdar" "random" and "separableNMF".
            See the initialization module for further details on each method.

        min_iterations: int
            The minimum number of iterations to perform during inference

        max_iterations: int
            The maximum number of iterations to perform during inference

        conv_test_freq: int
            The frequency at which the algorithm is tested for convergence.
            The objective function value is only computed every 'conv_test_freq'
            many iterations, which also affects a potentially saved history of
            the objective function values.

        tol: float
            The NMF algorithm is converged when the relative change
            of the objective function of one iteration is smaller
            than the tolerance 'tol'.
        """
        super().__init__(
            n_signatures,
            init_method,
            min_iterations,
            max_iterations,
            conv_test_freq,
            tol,
        )

        # initialize data/fitting dependent attributes
        self.W, self.H = None, None

    @property
    def signatures(self) -> pd.DataFrame:
        signatures = pd.DataFrame(
            self.W, index=self.mutation_types, columns=self.signature_names
        )
        return signatures

    @property
    def exposures(self) -> pd.DataFrame:
        exposures = pd.DataFrame(
            self.H, index=self.signature_names, columns=self.sample_names
        )
        return exposures

    @property
    def _n_parameters(self) -> int:
        """
        There are n_features * n_signatures parameters corresponding to
        the signature matrix and n_signatures * n_samples parameters
        corresponding to the exposure matrix.
        """
        return self.n_signatures * (self.n_features + self.n_samples)

    @abstractmethod
    def _update_W(self):
        pass

    @abstractmethod
    def _update_H(self):
        pass

    def _initialize(self, given_signatures=None, init_kwargs=None):
        """
        Initialize the signature matrix W and exposure matrix H.
        When the signatures are given, the initialization
        of W is overwritten by the given signatures.

        Input:
        ------
        given_signatures : pd.Dataframe, default=None
            At most 'n_signatures' many signatures can be provided to
            overwrite some of the initialized signatures. This does not
            change the initialized exposurse.

        init_kwargs: dict
            Any further keywords arguments to be passed to the initialization method.
            This includes, for example, a possible 'seed' keyword argument
            for all stochastic methods.
        """
        if given_signatures is not None:
            self._check_given_signatures(given_signatures)
            self.n_given_signatures = len(given_signatures.columns)
        else:
            self.n_given_signatures = 0

        init_kwargs = {} if init_kwargs is None else init_kwargs.copy()
        self.W, self.H, self.signature_names = initialize(
            self.X, self.n_signatures, self.init_method, given_signatures, **init_kwargs
        )

    @property
    def corr_signatures(self) -> pd.DataFrame:
        """
        The correlation of two signatures is given by the pearson correlation of
        the respective rows of the exposure matrix H.

        The pandas dataframe method 'corr' computes the pairwise correlation of columns.
        """
        return self.exposures.T.corr(method="pearson")

    @property
    def corr_samples(self) -> pd.DataFrame:
        """
        The correlation of two samples is given by the pearson correlation of
        the respective columns of the exposure matrix H.

        The pandas dataframe method 'corr' computes the pairwise correlation of columns.
        """
        return self.exposures.corr(method="pearson")

    def reorder(self, other_signatures, metric="cosine", keep_names=False):
        reordered_indices = match_signatures_pair(
            other_signatures, self.signatures, metric=metric
        )
        self.W = self.W[:, reordered_indices]
        self.H = self.H[reordered_indices, :]

        if keep_names:
            self.signature_names = self.signature_names[reordered_indices]

        return reordered_indices

    def _get_embedding_data(self):
        """
        In most NMF models like KL-NMF or mvNMF, the data for the embedding plot
        are just the (transposed) exposures.
        """
        return self.H.T.copy()

    def _get_default_embedding_annotations(self):
        """
        The embedding plot defaults to no annotations.
        """
        return None



import warnings

import fastcluster
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import umap
from adjustText import adjust_text
from scipy.cluster import hierarchy
from scipy.spatial.distance import pdist
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE

from .consts import COLORS_INDEL83, COLORS_SBS96, INDEL_TYPES_83, SBS_TYPES_96
from .utils import match_to_catalog, value_checker


def set_salamander_style():
    sns.set_context("notebook")
    sns.set_style("ticks")
    params = {
        "axes.edgecolor": "black",
        "axes.labelsize": 14,
        "axes.spines.top": False,
        "axes.spines.right": False,
        "axes.titlesize": 16,
        "errorbar.capsize": 3,
        "font.family": "DejaVu Sans",
        "legend.fontsize": 12,
        "lines.markersize": 8,
        "pdf.fonttype": 42,
        "xtick.labelsize": 12,
        "ytick.labelsize": 12,
    }
    mpl.rcParams.update(params)


def history_plot(values, conv_test_freq, min_iteration=0, ax=None, **kwargs):
    n_values = len(values)
    ns_iteration = np.arange(
        conv_test_freq, n_values * conv_test_freq + 1, conv_test_freq
    )

    if min_iteration > ns_iteration[-1]:
        raise ValueError(
            "The smallest iteration number shown in the history plot "
            "cannot be larger than the total number of iterations."
        )
    if ax is None:
        _, ax = plt.subplots(figsize=(4, 4))

    ax.set(xlabel="n_iteration", ylabel="objective function value")
    min_index = next(
        idx
        for idx, n_iteration in enumerate(ns_iteration)
        if n_iteration >= min_iteration
    )
    ax.plot(ns_iteration[min_index:], values[min_index:], **kwargs)
    return ax


def _annotate_plot(
    ax,
    data,
    annotations,
    ha="left",
    fontsize="medium",
    color="black",
    adjust_annotations=True,
    adjust_kwargs=None,
    **kwargs,
):
    for data_point, annotation in zip(data, annotations):
        ax.text(
            data_point[0],
            data_point[1],
            annotation,
            ha=ha,
            fontsize=fontsize,
            color=color,
            **kwargs,
        )
    if adjust_annotations:
        if adjust_kwargs is None:
            adjust_kwargs = {} if adjust_kwargs is None else adjust_kwargs.copy()

        annotations = [
            child for child in ax.get_children() if isinstance(child, mpl.text.Text)
        ]
        annotations = [
            annotation for annotation in annotations if annotation.get_text()
        ]
        adjust_text(annotations, **adjust_kwargs)


def _scatter_1d(data: np.ndarray, ax=None, **kwargs):
    if data.ndim != 1:
        raise ValueError(f"The datapoints of {data} (rows) have to be one-dimensional.")

    if ax is None:
        _, ax = plt.subplots(figsize=(6, 1))

    ax.spines[["left", "bottom"]].set_visible(False)
    ax.get_yaxis().set_visible(False)
    ax.axhline(y=0, color="black", zorder=1)
    data_2d = np.vstack([data, np.zeros_like(data)]).T
    sns.scatterplot(x=data_2d[:, 0], y=data_2d[:, 1], ax=ax, zorder=2, **kwargs)
    return data_2d, ax


def _scatter_2d(data, ax=None, **kwargs):
    """
    The rows (!) of 'data' are assumed to be the data points.
    """
    if data.shape[1] != 2:
        raise ValueError(f"The datapoints of {data} (rows) have to be two-dimensional.")

    if ax is None:
        _, ax = plt.subplots(figsize=(6, 6))

    ax.set(xlabel="x", ylabel="y")
    sns.scatterplot(x=data[:, 0], y=data[:, 1], ax=ax, **kwargs)
    return data, ax


def _pca_2d(data, ax=None, **kwargs):
    """
    The rows (!) of 'data' are assumed to be the data points.
    """
    if ax is None:
        _, ax = plt.subplots(figsize=(6, 6))

    data_2d = PCA(n_components=2).fit_transform(data)
    ax.set(xlabel="PC1", ylabel="PC2")
    sns.scatterplot(x=data_2d[:, 0], y=data_2d[:, 1], ax=ax, **kwargs)
    return data_2d, ax


def _tsne_2d(data, perplexity=30, ax=None, **kwargs):
    """
    The rows (!) of 'data' are assumed to be the single data points.
    """
    if ax is None:
        _, ax = plt.subplots(figsize=(6, 6))

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        perplexity = min(perplexity, len(data) - 1)
        data_2d = TSNE(perplexity=perplexity).fit_transform(data)

    ax.set(xlabel="t-SNE1", xticks=[], ylabel="t-SNE2", yticks=[])
    sns.scatterplot(x=data_2d[:, 0], y=data_2d[:, 1], ax=ax, **kwargs)
    return data_2d, ax


def _umap_2d(data, n_neighbors=15, min_dist=0.1, ax=None, **kwargs):
    """
    The rows (!) of 'data' are assumed to be the single data points.
    """
    if ax is None:
        _, ax = plt.subplots(figsize=(6, 6))

    n_neighbors = min(n_neighbors, len(data) - 1)
    data_2d = umap.UMAP(n_neighbors=n_neighbors, min_dist=min_dist).fit_transform(data)

    ax.set(xlabel="UMAP1", xticks=[], ylabel="UMAP2", yticks=[])
    sns.scatterplot(x=data_2d[:, 0], y=data_2d[:, 1], ax=ax, **kwargs)
    return data_2d, ax


def embeddings_plot(
    data: np.ndarray,
    method="umap",
    normalize=False,
    annotations=None,
    annotation_kwargs=None,
    adjust_annotations=True,
    adjust_kwargs=None,
    ax=None,
    **kwargs,
):
    """
    Plot a dimensionality reduction of high-dimensional data.
    If the dimension of the data is one or two, it is plotted
    directly, ignoring the chosen method.

    Parameters
    ----------
    data : np.ndarray
        The rows (!) are assumed to be the data points.

    method : str, default='umap'
        Either 'tsne', 'pca' or 'umap'. The respective dimensionality reduction
        is applied to plot the data in 2D. If the dimensionality of the data is
        less or equal to two, the points are plotted directly.

    normalize : bool, default=False
        If True, the data points are normalized to have a Euclidean norm of
        one before applying the dimensionality reduction.

    annotations : list[str], default=None
        Annotations of the projected data points. The length of the list has to match
        the number of data points.

    annotation_kwargs : dict, default=None
        keyword arguments to pass to matplotlibs plt.txt()

    adjust_annotations : bool, default=True
        If True, the annotations are adjusted to prevent them from overlapping.

    adjust_kwargs : dict, default=None
        keyword arguments to pass to adjustText's adjust_text.
        This is for example useful when arrows connecting the projected
        data points and the adjusted positions of the annotations
        should be added.

    ax : matplotlib.axes.Axes, default=None
        Pre-existing axes for the plot. Otherwise, an axes is created.

    **kwargs :
        keyword arguments to pass to seaborn's scatterplot

    Returns
    -------
    ax : matplotlib.axes.Axes
        The matplotlib axes containing the plot.
    """
    value_checker("method", method, ["pca", "tsne", "umap"])

    if normalize:
        data /= np.sqrt(np.sum(data**2, axis=1))[:, np.newaxis]

    n_dimensions = data.shape[1]

    if n_dimensions in [1, 2]:
        warnings.warn(
            f"The dimension of the data points is {n_dimensions}. "
            f"The method argument '{method}' will be ignored "
            "and the embeddings are plotted directly.",
            UserWarning,
        )

    if n_dimensions == 1:
        data_2d, ax = _scatter_1d(data[:, 0], ax=ax, **kwargs)
    elif n_dimensions == 2:
        data_2d, ax = _scatter_2d(data, ax=ax, **kwargs)
    elif method == "tsne":
        data_2d, ax = _tsne_2d(data, ax=ax, **kwargs)
    elif method == "pca":
        data_2d, ax = _pca_2d(data, ax=ax, **kwargs)
    else:
        data_2d, ax = _umap_2d(data, ax=ax, **kwargs)

    if annotations is not None:
        annotation_kwargs = (
            {} if annotation_kwargs is None else annotation_kwargs.copy()
        )
        _annotate_plot(
            ax,
            data_2d,
            annotations,
            adjust_annotations=adjust_annotations,
            adjust_kwargs=adjust_kwargs,
            **annotation_kwargs,
        )

    return ax


def corr_plot(
    corr: pd.DataFrame, figsize=(6, 6), cmap="vlag", annot=True, fmt=".2f", **kwargs
):
    linkage = hierarchy.linkage(corr)
    clustergrid = sns.clustermap(
        corr,
        row_linkage=linkage,
        figsize=figsize,
        vmin=-1,
        vmax=1,
        cmap=cmap,
        annot=annot,
        fmt=fmt,
        **kwargs,
    )

    return clustergrid


def _get_colors_signature_plot(mutation_types, colors=None):
    """
    Given the mutation types and the colors argument of sigplot_bar, return the
    final colors used in the signature bar chart.
    """
    n_features = len(mutation_types)

    if colors == "SBS96" or (
        n_features == 96 and all(mutation_types == SBS_TYPES_96) and colors is None
    ):
        if n_features != 96:
            raise ValueError(
                "The standard SBS colors can only be used "
                "when the signatures have 96 features."
            )
        colors = COLORS_SBS96

    elif colors == "Indel83" or (
        n_features == 83 and all(mutation_types == INDEL_TYPES_83) and colors is None
    ):
        if n_features != 83:
            raise ValueError(
                "The standard Indel colors can only be used "
                "when the signatures have 83 features."
            )
        colors = COLORS_INDEL83

    elif type(colors) in [str, tuple]:
        colors = n_features * [colors]

    elif type(colors) is list:
        if len(colors) != n_features:
            raise ValueError(
                f"The list of colors must be of length n_features={n_features}."
            )

    else:
        colors = n_features * ["gray"]

    return colors


def _signature_plot(
    signature, colors=None, annotate_mutation_types=False, ax=None, **kwargs
):
    """
    Inputs:
    -------
    signature: pd.Signature
        Signature with mutation types and name.

    colors: str, tuple or list
        Can be set to 'SBS96' or 'Indel83' to use the standard bar colors
        for these mutation types.
        Otherwise, when a single string or tuple is provided,
        all bars will have the same color. Alternatively,
        a list can be used to specifiy the color of each bar individually.

    ax:
        A single matplotlib Axes in which to draw the plot.

    kwargs: dict
        Any keyword arguments to be passed to matplotlibs ax.bar
    """
    if ax is None:
        _, ax = plt.subplots(figsize=(4, 1))

    signature_normalized = signature / signature.sum(axis=0)
    mutation_types = signature.index
    colors = _get_colors_signature_plot(mutation_types, colors)

    ax.set_title(signature_normalized.columns[0])
    ax.spines["left"].set_visible(False)
    ax.get_yaxis().set_visible(False)
    ax.set_xlim((-1, len(mutation_types)))

    ax.bar(
        mutation_types,
        signature_normalized.iloc[:, 0],
        linewidth=0,
        color=colors,
        **kwargs,
    )

    if annotate_mutation_types:
        ax.set_xticks(mutation_types)
        ax.set_xticklabels(
            mutation_types, family="monospace", fontsize=4, ha="center", rotation=90
        )

    else:
        ax.set_xticks([])

    return ax


def signature_plot(
    signature,
    catalog=None,
    colors=None,
    annotate_mutation_types=False,
    ax=None,
    **kwargs,
):
    """
    Inputs:
    -------
    signature: pd.Signature
        Signature with mutation types and name.

    catalog: pd.DataFrame
        If a catalog is provided, the single best matching catalog signature
        will also be plotted.

    colors: str, tuple or list
        Can be set to 'SBS96' or 'Indel83' to use the standard bar colors
        for these mutation types.
        Otherwise, when a single string or tuple is provided,
        all bars will have the same color. Alternatively,
        a list can be used to specifiy the color of each bar individually.

    ax:
        Axes in which to draw the plot. A single Axes if catalog is None;
        two Axes if a catalog is given.

    kwargs: dict
        Any keyword arguments to be passed to matplotlibs ax.bar
    """
    if catalog is None:
        if ax is None:
            _, ax = plt.subplots(figsize=(4, 1))

        signatures = [signature]
        axes = [ax]

    else:
        if ax is None:
            _, ax = plt.subplots(1, 2, figsize=(8, 1))

        matched_signature = match_to_catalog(signature, catalog, metric="cosine")
        signatures = [signature, matched_signature]
        axes = ax

    for sig, axis in zip(signatures, axes):
        _signature_plot(
            sig,
            colors=colors,
            annotate_mutation_types=annotate_mutation_types,
            ax=axis,
            **kwargs,
        )

    if catalog is None:
        return axes[0]

    return axes


def signatures_plot(
    signatures,
    catalog=None,
    colors=None,
    annotate_mutation_types=False,
    axes=None,
    **kwargs,
):
    """
    Inputs:
    -------
    signatures : pd.DataFrame
        Named signatures of shape (n_features, n_signatures)

    catalog: pd.DataFrame
        If a catalog is provided, the best matching catalog signatures
        will also be plotted.

    axes : list
        Axes in which to draw the plot. Multiple Axes if more than one signature
        is provided or a catalog is given. Otherwise a single axis.
        When a catalog is provided, axes is expected to be of shape (n_signatures, 2).
    """
    n_signatures = signatures.shape[1]

    if n_signatures == 1:
        ax = signature_plot(
            signatures,
            catalog=catalog,
            colors=colors,
            annotate_mutation_types=annotate_mutation_types,
            ax=axes,
            **kwargs,
        )
        return ax

    if axes is None:
        if catalog is None:
            _, axes = plt.subplots(n_signatures, 1, figsize=(4, n_signatures))
        else:
            _, axes = plt.subplots(n_signatures, 2, figsize=(8, n_signatures))

    for ax, signature in zip(axes, signatures):
        signature_plot(
            signatures[[signature]],
            catalog=catalog,
            colors=colors,
            annotate_mutation_types=annotate_mutation_types,
            ax=ax,
            **kwargs,
        )
    plt.tight_layout()

    return axes


def _get_sample_order(exposures: pd.DataFrame, normalize=True):
    """
    Compute the aesthetically most pleasing order of the samples
    for a stacked bar chart of the exposures.

    Parameters
    ----------
    exposures : pd.DataFrame of shape (n_signatures, n_samples)
        The named exposure matrix

    normalize : bool, default=True
        If True, the exposures are normalized before computing the
        hierarchical clustering.

    Returns
    -------
    sample_order : np.ndarray
        The ordered sample names
    """
    if normalize:
        # not in-place
        exposures = exposures / exposures.sum(axis=0)

    d = pdist(exposures.T)
    linkage = fastcluster.linkage(d)
    # get the optimal sample order that is consistent
    # with the hierarchical clustering linkage
    sample_order = hierarchy.leaves_list(hierarchy.optimal_leaf_ordering(linkage, d))
    sample_order = exposures.columns[sample_order].to_numpy()
    return sample_order


def _reorder_exposures(
    exposures: pd.DataFrame, sample_order=None, reorder_signatures=True
):
    """
    Reorder the samples with hierarchical clustering and
    reorder the signatures by their total relative exposure.

    Parameters
    ----------
    exposures : pd.DataFrame of shape (n_signatures, n_samples)
        The named exposure matrix

    sample_order : np.ndarray, default=None
        A predefined order of the samples as a list of sample names.
        If None, hierarchical clustering is used to compute the
        aesthetically most pleasing order.

    reorder_signatures : bool, default=True
        If True, the signatures will be reordered such that the
        total relative exposures of the signatures decrease from the bottom
        to the top signature in the stacked bar chart.

    Returns
    -------
    exposures_reordered : pd.DataFrame of shape (n_signatures, n_samples)
        The reorderd named exposure matrix
    """
    if sample_order is None:
        sample_order = _get_sample_order(exposures)

    exposures_reordered = exposures[sample_order]

    # order the signatures by their total relative exposure
    if reorder_signatures:
        exposures_normalized = exposures_reordered / exposures_reordered.sum(axis=0)
        signature_order = (
            exposures_normalized.sum(axis=1).sort_values(ascending=False).index
        )
        exposures_reordered = exposures_reordered.reindex(signature_order)

    return exposures_reordered


def exposures_plot(
    exposures: pd.DataFrame,
    sample_order=None,
    reorder_signatures=True,
    annotate_samples=True,
    colors=None,
    ncol_legend=1,
    ax=None,
    **kwargs,
):
    """
    Visualize the exposures with a stacked bar chart.

    Parameter
    ---------
    exposures : pd.DataFrame of shape (n_signatures, n_samples)
        The named exposure matrix.

    sample_order : np.ndarray, default=None
        A predefined order of the samples as a list of sample names.
        If None, hierarchical clustering is used to compute the
        aesthetically most pleasing order.

    reorder_signatures : bool, default=True
        If True, the signatures will be reordered such that the
        total relative exposures of the signatures decrease from the bottom
        to the top signature in the stacked bar chart.

    annotate_samples : bool, default=True
        If True, the x-axis is annotated with the sample names.

    colors : list of length n_signatures, default=None
        Colors to pass to matplotlibs ax.bar, one per signature.

    n_col_legend : int, default=1
        The number of columns of the legend.

    ax : matplotlib.axes.Axes, default=None
        Pre-existing axes for the plot. Otherwise, create an axis internally.

    kwargs : dict
        Any keyword arguments to be passed to matplotlibs ax.bar.

    Returns
    -------
    ax : matplotlib.axes.Axes
        The matplotlib axes containing the plot.
    """
    n_signatures, n_samples = exposures.shape
    # not in-place
    exposures = exposures / exposures.sum(axis=0)
    exposures = _reorder_exposures(
        exposures, sample_order=sample_order, reorder_signatures=reorder_signatures
    )

    if ax is None:
        _, ax = plt.subplots(figsize=(0.3 * n_samples, 4))

    if colors is None:
        colors = list(sns.color_palette("deep")) * (1 + n_signatures // 10)

    bottom = np.zeros(n_samples)

    for signature, color in zip(exposures.T, colors):
        signature_exposures = exposures.T[signature].to_numpy()
        ax.bar(
            np.arange(n_samples),
            signature_exposures,
            color=color,
            width=1,
            label=signature,
            linewidth=0,
            bottom=bottom,
            **kwargs,
        )
        bottom += signature_exposures

    if annotate_samples:
        ax.set_xticks(np.arange(n_samples))
        ax.set_xticklabels(exposures.columns, rotation=90, ha="center", fontsize=10)

    else:
        ax.get_xaxis().set_visible(False)

    ax.set_title("Sample exposures")
    ax.spines[["left", "bottom"]].set_visible(False)
    ax.get_yaxis().set_visible(False)
    ax.legend(loc="center left", bbox_to_anchor=(0.975, 0.5), ncol=ncol_legend)

    return ax



"""
Initialization methods for non-negative matrix factorization (NMF)
"""
import numpy as np
from sklearn.decomposition import _nmf as sknmf

from ..utils import normalize_WH, shape_checker, type_checker, value_checker

EPSILON = np.finfo(np.float32).eps
INIT_METHODS = [
    "custom",
    "flat",
    "hierarchical_cluster",
    "nndsvd",
    "nndsvda",
    "nndsvdar",
    "random",
    "separableNMF",
]


def init_custom(
    X: np.ndarray, n_signatures: int, W_custom: np.ndarray, H_custom: np.ndarray
):
    """
    Perform type and shape checks on custom signature and
    exposure matrix initializations.
    """
    type_checker("W_custom", W_custom, np.ndarray)
    type_checker("H_custom", H_custom, np.ndarray)
    n_features, n_samples = X.shape
    shape_checker("W_custom", W_custom, (n_features, n_signatures))
    shape_checker("H_custom", H_custom, (n_signatures, n_samples))
    return W_custom, H_custom


def init_flat(X: np.ndarray, n_signatures: int):
    """
    Initialize the signature and exposure matrices with one float, respectively.
    """
    n_features, n_samples = X.shape
    scaling = np.mean(np.sum(X, axis=0))
    W = np.full((n_features, n_signatures), 1 / n_features)
    H = np.full((n_signatures, n_samples), scaling / n_signatures)
    return W, H


def init_nndsvd(X: np.ndarray, n_signatures: int, init: str, seed=None):
    """
    A wrapper around the non-negative double singular value decomposition (NNDSVD)
    initialization methods "nndsvd", "nndsvda" and "nndsvdar" from scikit-learn.

    Inputs:
    ------
    init: str
        One of "nndsvd", "nndsvda" and "nndsvdar"
    """
    if seed is not None:
        np.random.seed(seed)

    # pylint: disable-next=W0212
    W, H = sknmf._initialize_nmf(X, n_signatures, init=init)

    return W, H


def init_random(X: np.ndarray, n_signatures: int, seed=None):
    """
    Initialize each signature by drawing from the uniform
    distribution on the simplex.
    Initialize the exposures of each sample as a scaled sample
    from the uniform distribution on a simplex.
    The scaling is chosen such that the expected total exposure is equal to
    the column sum of that sample in the count matrix X.
    """
    if seed is not None:
        np.random.seed(seed)

    n_features, n_samples = X.shape
    W = np.random.dirichlet(np.ones(n_features), size=n_signatures).T
    scaling = np.sum(X, axis=0)
    H = scaling * np.random.dirichlet(np.ones(n_signatures), size=n_samples).T
    return W, H


def init_separableNMF(X: np.ndarray, n_signatures: int, seed=None):
    r"""
    This code is following Algorithm 1 from "Fast and Robust Recursive
    Algorithms for Separable Nonnegative Matrix Factorization"
    (Gillis and Vavasis, 2013), with the canonical choice of
    f(x) = \| x \|_2^2 as the strongly convex function f satisfying
    Assumption 2 from the paper.
    """
    signature_indices = np.empty(n_signatures, dtype=int)
    R = X / np.sum(X, axis=0)

    for k in range(n_signatures):
        column_norms = np.sum(R**2, axis=0)
        kstar = np.argmax(column_norms)
        u = R[:, kstar]
        R = (np.identity(X.shape[0]) - np.outer(u, u) / column_norms[kstar]) @ R
        signature_indices[k] = kstar

    W = X[:, signature_indices].astype(float)
    _, H = init_random(X, n_signatures, seed=seed)
    return W, H


def initialize(
    X: np.ndarray,
    n_signatures: int,
    init_method="nndsvd",
    given_signatures=None,
    **kwargs,
):
    """
    Initialize the signature and exposure matrices.

    Parameters
    ----------
    X : np.ndarray
        count matrix

    n_signatures : int
        number of signatures

    init_method : str
        initialization method. One of 'custom', 'flat', 'hierarchical_cluster',
        'nndsvd', 'nndsvda', 'nndsvdar', 'random', 'separableNMF'

    given_signatures : pd.Dataframe, default=None
        At most 'n_signatures' many signatures can be provided to
        overwrite some of the initialized signatures. This does not
        change the initialized exposurse.

    kwargs : dict
        Any keyword arguments to be passed to the initialization method.
        This includes, for example, a possible 'seed' keyword argument
        for all stochastic methods.

    Returns
    -------
    W : np.ndarray
        signature matrix

    H : np.ndarray
        exposure matrix

    signature_names : list
        The signature names. By default, the signatures are named
        'Sigk', where 'k' is one plus the index of the signature.
        If 'given_signatures' are provided, the names are adjusted
        accordingly.
    """
    value_checker("init_method", init_method, INIT_METHODS)

    if init_method == "custom":
        W, H = init_custom(X, n_signatures, **kwargs)

    elif init_method == "flat":
        W, H = init_flat(X, n_signatures)

    elif init_method in ["nndsvd", "nndsvda", "nndsvdar"]:
        W, H = init_nndsvd(X, n_signatures, init=init_method, **kwargs)

    elif init_method == "random":
        W, H = init_random(X, n_signatures, **kwargs)

    else:
        W, H = init_separableNMF(X, n_signatures, **kwargs)

    if given_signatures is not None:
        n_given_signatures = len(given_signatures.columns)
        W[:, :n_given_signatures] = given_signatures.copy().values
        given_signatures_names = given_signatures.columns.to_numpy(dtype=str)
        n_new_signatures = n_signatures - n_given_signatures
        new_signatures_names = np.array([f"Sig{k+1}" for k in range(n_new_signatures)])
        signature_names = np.concatenate([given_signatures_names, new_signatures_names])
    else:
        signature_names = np.array([f"Sig{k+1}" for k in range(n_signatures)])

    W, H = normalize_WH(W, H)
    W, H = W.clip(EPSILON), H.clip(EPSILON)
    return W, H, signature_names



import numpy as np
import pandas as pd
from numba import njit
from scipy.optimize import linear_sum_assignment
from sklearn.metrics import pairwise_distances

EPSILON = np.finfo(np.float32).eps


def shape_checker(arg_name: str, arg, allowed_shape):
    """
    A helper function to test the shape of a numpy ndarray or pandas dataframe.

    Input:
    ------
    arg_name: str
        The name of the argument
    arg:
        The actual value of the argument
    allowed_shape:
        The expected shape of 'arg'
    """
    type_checker(arg_name, arg, [np.ndarray, pd.DataFrame])

    if arg.shape != allowed_shape:
        raise ValueError(f"The shape of '{arg_name}' has to be {allowed_shape}.")


def type_checker(arg_name: str, arg, allowed_types):
    """
    A helper function to test the type of an argument.

    Input:
    ------
    arg_name: str
        The name of the argument
    arg:
        The actual value of the argument
    allowed_types: a type or list of types
        The type or list of types allowed for 'arg'
    """
    if isinstance(allowed_types, type):
        allowed_types = [allowed_types]

    if type(arg) not in allowed_types:
        raise TypeError(f"The type of '{arg_name}' has to be one of {allowed_types}.")


def value_checker(arg_name: str, arg, allowed_values):
    """
    A helper function to test the value of an argument.

    Input:
    ------
    arg_name: str
        The name of the argument
    arg:
        The actual value of the argument
    allowed_values:
        A value or list of values allowed for 'arg'
    """
    if not isinstance(allowed_values, list):
        allowed_values = [allowed_values]

    if arg not in allowed_values:
        raise ValueError(
            f"The value of '{arg_name}' has to be one of {allowed_values}."
        )


@njit
def normalize_WH(W, H):
    normalization_factor = np.sum(W, axis=0)
    return W / normalization_factor, H * normalization_factor[:, None]


def match_to_catalog(signatures: pd.DataFrame, catalog: pd.DataFrame, metric="cosine"):
    """
    Find the best matching signatures in catalog for all signatures.
    """
    cosine_sim = 1 - pairwise_distances(signatures.T, catalog.T, metric=metric)
    matches_indices = [np.argmax(row) for row in cosine_sim]
    matches = catalog.iloc[:, matches_indices]

    return matches


def match_signatures_pair(
    signatures1: pd.DataFrame, signatures2: pd.DataFrame, metric="cosine"
):
    """
    Match a pair of signature catalogs using their pairwise column distances,
    see https://en.wikipedia.org/wiki/Assignment_problem.

    Output:
    ------
    reordered_indices: np.ndarray
        The list of column indices such that reordering signatures2 using this list
        minimizes the sum of the pairwise column distances between
        signatures1 and signatures2.
    """
    if signatures1.shape != signatures2.shape:
        raise ValueError("The signatures must be of the same shape.")

    pdist = pairwise_distances(signatures1.T, signatures2.T, metric=metric)
    reordered_indices = linear_sum_assignment(pdist)[1]

    return reordered_indices



# This implementation relies on helper functions in corrnmf.py.
# In particular, functions with leading '_'' are accessed
# pylint: disable=protected-access

import numpy as np
import pandas as pd
from scipy import optimize

from . import _utils_corrnmf
from ._utils_klnmf import update_W
from .corrnmf import CorrNMF

EPSILON = np.finfo(np.float32).eps


class CorrNMFDet(CorrNMF):
    r"""
    The CorrNMFDet class implements the deterministic batch version of
    a variant of the correlated NMF (CorrNMF) algorithm devolped in
    "Bayesian Nonnegative Matrix Factorization with Stochastic Variational
    Inference" by Paisley et al.

    The following methods are implemented to match the structure
    of the abstract class CorrNMF:

        - _update_alpha:
            update the sample exposure biases \alpha

        - _update_beta:
            update the signature biases \beta

        - _update_sigma_sq:
            update the variance \sigma^2 assumed in the generative model

        - _update_W:
            update the signature matrix W

        - _update_p:
            update the auxiliary parameters p

        - _update_l:
            update a single signature embedding l

        - _update_u:
            update a single sample embedding u

    The following method is implemented to match the structure of SignatureNMF:

        - fit:
            Perform CorrNMF for the given mutation count data or
            for given signatures and mutation count data
    """

    def _update_alpha(self, given_sample_biases=None):
        if given_sample_biases is None:
            self.alpha = _utils_corrnmf.update_alpha(self.X, self.beta, self.L, self.U)

    def _update_beta(self, p, given_signature_biases=None):
        if given_signature_biases is None:
            self.beta = _utils_corrnmf.update_beta(
                self.X, p, self.alpha, self.L, self.U
            )

    def _update_sigma_sq(self):
        embeddings = np.concatenate([self.L, self.U], axis=1)
        self.sigma_sq = np.mean(embeddings**2)
        self.sigma_sq = np.clip(self.sigma_sq, EPSILON, None)

    def _update_W(self):
        self.W = update_W(
            self.X,
            self.W,
            self.exposures.values,
            n_given_signatures=self.n_given_signatures,
        )

    def _update_p(self):
        p = _utils_corrnmf.update_p_unnormalized(self.W, self.exposures.values)
        p /= np.sum(p, axis=1, keepdims=True)
        p = p.clip(EPSILON)
        return p

    def _update_l(self, index, aux_row, outer_prods_U):
        beta = self.beta[index]

        def objective_fun(l):
            return _utils_corrnmf.objective_function_embedding(
                l, self.U, self.alpha, beta, self.sigma_sq, aux_row
            )

        summand_grad = np.sum(aux_row * self.U, axis=1)

        def gradient(l):
            return _utils_corrnmf.gradient_embedding(
                l, self.U, self.alpha, beta, self.sigma_sq, summand_grad
            )

        def hessian(l):
            return _utils_corrnmf.hessian_embedding(
                l, self.U, self.alpha, beta, self.sigma_sq, outer_prods_U
            )

        l = optimize.minimize(
            fun=objective_fun,
            x0=self.L[:, index],
            method="Newton-CG",
            jac=gradient,
            hess=hessian,
        ).x
        l[(0 < l) & (l < EPSILON)] = EPSILON
        l[(-EPSILON < l) & (l < 0)] = -EPSILON
        self.L[:, index] = l

    def _update_L(self, aux, outer_prods_U=None):
        r"""
        Update all signature embeddings by optimizing
        the surrogate objective function using scipy.optimize.minimize
        with the 'Newton-CG' method (strictly convex for each embedding).

        aux: np.ndarray
            aux_kd = \sum_v X_vd * p_vkd
            is used for updating the signatures and the sample embeddidngs.
        """
        if outer_prods_U is None:
            outer_prods_U = np.einsum("mD,nD->Dmn", self.U, self.U)

        for k, aux_row in enumerate(aux):
            self._update_l(k, aux_row, outer_prods_U)

    def _update_u(self, index, aux_col, outer_prods_L):
        alpha = self.alpha[index]

        def objective_fun(u):
            return _utils_corrnmf.objective_function_embedding(
                u, self.L, alpha, self.beta, self.sigma_sq, aux_col
            )

        summand_grad = np.sum(aux_col * self.L, axis=1)

        def gradient(u):
            return _utils_corrnmf.gradient_embedding(
                u, self.L, alpha, self.beta, self.sigma_sq, summand_grad
            )

        def hessian(u):
            return _utils_corrnmf.hessian_embedding(
                u, self.L, alpha, self.beta, self.sigma_sq, outer_prods_L
            )

        u = optimize.minimize(
            fun=objective_fun,
            x0=self.U[:, index],
            method="Newton-CG",
            jac=gradient,
            hess=hessian,
            options={"maxiter": 3},
        ).x
        u[(0 < u) & (u < EPSILON)] = EPSILON
        u[(-EPSILON < u) & (u < 0)] = -EPSILON
        self.U[:, index] = u

    def _update_U(self, aux):
        r"""
        Update all sample embeddings by optimizing
        the surrogate objective function using scipy.optimize.minimize
        with the 'Newton-CG' method (strictly convex for each embedding).

        aux: np.ndarray
            aux_kd = \sum_v X_vd * p_vkd
            is used for updating the signatures and the sample embeddidngs.
        """
        outer_prods_L = np.einsum("mK,nK->Kmn", self.L, self.L)

        for d, aux_col in enumerate(aux.T):
            self._update_u(d, aux_col, outer_prods_L)

    def _update_LU(
        self, p, given_signature_embeddings=None, given_sample_embeddings=None
    ):
        aux = np.einsum("vd,vkd->kd", self.X, p)

        if given_signature_embeddings is None:
            self._update_L(aux)

        if given_sample_embeddings is None:
            self._update_U(aux)

    def fit(
        self,
        data: pd.DataFrame,
        given_signatures=None,
        given_signature_biases=None,
        given_signature_embeddings=None,
        given_sample_biases=None,
        given_sample_embeddings=None,
        init_kwargs=None,
        history=False,
        verbose=0,
    ):
        """
        Maximize the surrogate objective function of correlated NMF (CNMF).

        Input:
        ------
        data: pd.DataFrame
            The mutation count data

        given_signatures: pd.DataFrame, default=None
            Known signatures that will be fixed during model fitting.

        given_signature_biases : np.ndarray, default=None
            Known signature biases of shape (n_signatures,) that will be fixed
            during model fitting.

        given_signature_embeddings: np.ndarray, default=None
            Known signature embeddings that will be fixed during model fitting.

        given_sample_biases : np.ndarray, default=None
            Known sample biases of shape (n_samples,) that will be fixed
            during model fitting.

        given_sample_embeddings: np.ndarray, default=None
            Known sample embeddings that will be fixed during model fitting.

        init_kwargs: dict
            Any further keywords arguments to be passed to the initialization method.
            This includes, for example, a possible 'seed' keyword argument
            for all stochastic methods.

        history: bool
            When set to true, the history of the objective function and
            surrogate objective function will be stored in a dictionary.

        verbose: int
            Every 100th iteration number will be printed when set unequal to zero.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        self._setup_data_parameters(data)
        self._initialize(
            given_signatures=given_signatures,
            given_signature_biases=given_signature_biases,
            given_signature_embeddings=given_signature_embeddings,
            given_sample_biases=given_sample_biases,
            given_sample_embeddings=given_sample_embeddings,
            init_kwargs=init_kwargs,
        )
        of_values = [self.objective_function()]
        n_iteration = 0
        converged = False

        while not converged:
            n_iteration += 1

            if verbose and n_iteration % 100 == 0:
                print(f"iteration: {n_iteration}; objective: {of_values[-1]:.2f}")

            self._update_alpha(given_sample_biases)
            p = self._update_p()
            self._update_beta(p, given_signature_biases)
            self._update_LU(p, given_signature_embeddings, given_sample_embeddings)
            self._update_sigma_sq()

            if self.n_given_signatures < self.n_signatures:
                self._update_W()

            if n_iteration % self.conv_test_freq == 0:
                prev_of_value = of_values[-1]
                of_values.append(self.objective_function())
                rel_change = (of_values[-1] - prev_of_value) / np.abs(prev_of_value)
                converged = rel_change < self.tol and n_iteration >= self.min_iterations

            converged |= n_iteration >= self.max_iterations

        if history:
            self.history["objective_function"] = of_values[1:]

        return self



import numpy as np
from numba import njit
from scipy.special import gammaln

EPSILON = np.finfo(np.float32).eps


@njit(fastmath=True)
def kl_divergence(X: np.ndarray, W: np.ndarray, H: np.ndarray, weights=None) -> float:
    r"""
    The generalized Kullback-Leibler divergence
    D_KL(X || WH) = \sum_vd X_vd * ln(X_vd / (WH)_vd) - \sum_vd X_vd + \sum_vd (WH)_vd.

    Parameters
    ----------
    X : np.ndarray of shape (n_features, n_samples)
        data matrix

    W : np.ndarray of shape (n_features, n_signatures)
        signature matrix

    H : np.ndarray of shape (n_signatures, n_samples)
        exposure matrix

    weights : np.ndarray of shape (n_samples,)
        per sample weights

    Returns
    -------
    result : float
    """
    V, D = X.shape
    WH = W @ H
    result = 0.0

    for d in range(D):
        summand_sample = 0.0

        for v in range(V):
            if X[v, d] != 0:
                summand_sample += X[v, d] * np.log(X[v, d] / WH[v, d])
                summand_sample -= X[v, d]
            summand_sample += WH[v, d]

        if weights is not None:
            summand_sample *= weights[d]

        result += summand_sample

    return result


def samplewise_kl_divergence(
    X: np.ndarray, W: np.ndarray, H: np.ndarray, weights=None
) -> np.ndarray:
    """
    Per sample (weighted) generalized Kullback-Leibler divergence D_KL(x || Wh).

    Parameters
    ----------
    X : np.ndarray of shape (n_features, n_samples)
        data matrix

    W : np.ndarray of shape (n_features, n_signatures)
        signature matrix

    H : np.ndarray of shape (n_signatures, n_samples)
        exposure matrix

    weights : np.ndarray of shape (n_samples,)
        per sample weights

    Returns
    -------
    errors : np.ndarray of shape (n_samples,)
    """
    X_data = np.copy(X).astype(float)
    indices = X == 0
    X_data[indices] = EPSILON
    WH_data = W @ H
    WH_data[indices] = EPSILON

    s1 = np.einsum("vd,vd->d", X_data, np.log(X_data / WH_data))
    s2 = -np.sum(X, axis=0)
    s3 = np.dot(H.T, np.sum(W, axis=0))

    errors = s1 + s2 + s3

    if weights is not None:
        errors *= weights

    return errors


@njit(fastmath=True)
def _poisson_llh_wo_factorial(X: np.ndarray, W: np.ndarray, H: np.ndarray) -> float:
    """
    The Poisson log-likelihood generalized to X, W and H having
    non-negative real numbers without the summands involving the log-factorial
    of elements of X.
    Note:
        scipy-special, which is required to computed the log-factorial,
        is not supported by numba.

    Parameters
    ----------
    X : np.ndarray of shape (n_features, n_samples)
        data matrix

    W : np.ndarray of shape (n_features, n_signatures)
        signature matrix

    H : np.ndarray of shape (n_signatures, n_samples)
        exposure matrix

    Returns
    -------
    result : float
    """
    V, D = X.shape
    WH = W @ H
    result = 0.0

    for v in range(V):
        for d in range(D):
            if WH[v, d] != 0:
                result += X[v, d] * np.log(WH[v, d])
            result -= WH[v, d]

    return result


def poisson_llh(X: np.ndarray, W: np.ndarray, H: np.ndarray) -> float:
    """
    The Poisson log-likelihood generalized to X, W and H having
    non-negative real numbers.

    Parameters
    ----------
    X : np.ndarray of shape (n_features, n_samples)
        data matrix

    W : np.ndarray of shape (n_features, n_signatures)
        signature matrix

    H : np.ndarray of shape (n_signatures, n_samples)
        exposure matrix

    Returns
    -------
    result : float
    """
    result = _poisson_llh_wo_factorial(X, W, H)
    result -= np.sum(gammaln(1 + X))

    return result


@njit
def update_W(
    X: np.ndarray,
    W: np.ndarray,
    H: np.ndarray,
    weights_kl=None,
    n_given_signatures: int = 0,
) -> np.ndarray:
    """
    The multiplicative update rule of the signature matrix W
    under the constraint of normalized signatures. It can be shown
    that the generalized KL-divegence D_KL(X || WH) is decreasing
    under the implemented update rule.

    Clipping the matrix avoids floating point errors.

    Parameters
    ----------
    X : np.ndarray of shape (n_features, n_samples)
        data matrix

    W : np.ndarray of shape (n_features, n_signatures)
        signature matrix

    H : np.ndarray of shape (n_signatures, n_samples)
        exposure matrix

    weights_kl : np.ndarray of shape (n_samples,)
        per sample weights in the KL-divergence loss

    n_given_signatures : int
        The number of known signatures which will not be updated.

    Returns
    -------
    W : np.ndarray of shape (n_features, n_signatures)
        updated signature matrix
    """
    n_signatures = W.shape[1]

    if n_given_signatures == n_signatures:
        return W

    aux = X / (W @ H)

    if weights_kl is not None:
        aux *= weights_kl

    W_updated = W * (aux @ H.T)
    W_updated /= W_updated.sum(axis=0)
    W_updated[:, :n_given_signatures] = W[:, :n_given_signatures].copy()
    W_updated[:, n_given_signatures:] = W_updated[:, n_given_signatures:].clip(EPSILON)

    return W_updated


@njit
def update_H(
    X: np.ndarray, W: np.ndarray, H: np.ndarray, weights_kl=None, weights_l_half=None
) -> np.ndarray:
    """
    The multiplicative update rule of the exposure matrix H
    under the constraint of normalized signatures.

    Clipping the matrix avoids floating point errors.

    Parameters
    ----------
    X : np.ndarray of shape (n_features, n_samples)
        data matrix

    W : np.ndarray of shape (n_features, n_signatures)
        signature matrix

    H : np.ndarray of shape (n_signatures, n_samples)
        exposure matrix

    weights_kl : np.ndarray of shape (n_samples,)
        per sample weights in the KL-divergence loss

    weights_l_half : np.ndarray of shape (n_samples,)
        per sample l_half penalty weights. They can be used to induce
        sparse exposures.

    Returns
    -------
    H_updated : np.ndarray of shape (n_signatures, n_samples)
        The updated exposure matrix. If possible, the update is performed
        in-place.
    """
    aux = X / (W @ H)

    if weights_l_half is None:
        # in-place
        H *= W.T @ aux
        H = H.clip(EPSILON)
        return H

    intermediate = 4.0 * H * (W.T @ aux)

    if weights_kl is not None:
        intermediate *= weights_kl**2

    discriminant = 0.25 * weights_l_half**2 + intermediate
    H_updated = 0.25 * (weights_l_half / 2 - np.sqrt(discriminant)) ** 2

    if weights_kl is not None:
        H_updated /= weights_kl**2

    H_updated = H_updated.clip(EPSILON)
    return H_updated


@njit
def update_WH(
    X: np.ndarray,
    W: np.ndarray,
    H: np.ndarray,
    weights_kl=None,
    weights_l_half=None,
    n_given_signatures: int = 0,
) -> np.ndarray:
    """
    A joint update rule for the signature matrix W and
    the exposure matrix H under the constraint of normalized
    signatures.

    Clipping the matrix avoids floating point errors.

    Parameters
    ----------
    X : np.ndarray of shape (n_features, n_samples)
        data matrix

    W : np.ndarray of shape (n_features, n_signatures)
        signature matrix

    H : np.ndarray of shape (n_signatures, n_samples)
        exposure matrix

    weights_kl : np.ndarray of shape (n_samples,)
        per sample weights in the KL-divergence loss

    weights_l_half : np.ndarray of shape (n_samples,)
        per sample l_half penalty weights. They can be used to induce
        sparse exposures.

    n_given_signatures : int
        The number of known signatures which will not be updated.

    Returns
    -------
    W_updated : np.ndarray of shape (n_features, n_signatures)
        updated signature matrix

    H_updated : np.ndarray of shape (n_signatures, n_samples)
        The updated exposure matrix. If possible, the update is performed
        in-place.
    """
    n_signatures = W.shape[1]
    aux = X / (W @ H)

    if n_given_signatures == n_signatures:
        W_updated = W
    else:
        if weights_kl is None:
            scaled_aux = aux
        else:
            scaled_aux = weights_kl * aux
        # the old signatures are needed for updating H
        W_updated = W * (scaled_aux @ H.T)
        W_updated /= np.sum(W_updated, axis=0)
        W_updated[:, :n_given_signatures] = W[:, :n_given_signatures].copy()
        W_updated = W_updated.clip(EPSILON)

    if weights_l_half is None:
        # in-place
        H *= W.T @ aux
        H = H.clip(EPSILON)
        return W_updated, H

    intermediate = 4.0 * H * (W.T @ aux)

    if weights_kl is not None:
        intermediate *= weights_kl**2

    discriminant = 0.25 * weights_l_half**2 + intermediate
    H_updated = 0.25 * (weights_l_half / 2 - np.sqrt(discriminant)) ** 2

    if weights_kl is not None:
        H_updated /= weights_kl**2

    H_updated = H_updated.clip(EPSILON)
    return W_updated, H_updated



from abc import abstractmethod

import numpy as np
import pandas as pd
from scipy.spatial.distance import squareform
from scipy.special import gammaln

from ..utils import match_signatures_pair, shape_checker, type_checker
from ._utils_klnmf import kl_divergence, poisson_llh, samplewise_kl_divergence
from .initialization import initialize
from .signature_nmf import SignatureNMF

EPSILON = np.finfo(np.float32).eps


class CorrNMF(SignatureNMF):
    r"""
    The abstract class CorrNMF unifies the structure of deterministic and
    stochastic correlated NMF (CorrNMF) with and without given signatures.
    Both variants of CorrNMF have an identical generative model and objective function.
    The model parameters are the sample biases \alpha, variance \sigma^2,
    signature matrix W and the auxiliary parameters p.
    The latent variables are the signature embeddings L and the sample embeddings U.

    Overview:

    Every child class has to implement the following methods:

        - _update_alpha:
            update the sample exposure biases \alpha

        - _update_beta:
            update the signature exposure biases \beta

        - _update_sigma_sq:
            update the embedding distribution variance \sigma^2

        - _update_W:
            update the signature matrix W

        - _update_p:
            update the auxiliary parameters p

        - _update_l:
            update a single signature embedding l

        - _update_u:
            update a single sample embedding u

        - fit:
            Run CorrNMF for a given mutation count data. Every
            fit method should also implement a version that allows fixing
            arbitrary many a priori known signatures.


    The following attributes are implemented in the abstract class CNMF:

        - signatures: pd.DataFrame
            The signature matrix including mutation type names and signature names

        - exposures: pd.DataFrame
            The exposure matrix including the signature names and sample names

        - reconstruction_error: float
            The reconstruction error between the count matrix
            and the reconstructed count matrix.

        - samplewise_reconstruction_error: np.ndarray
            The samplewise reconstruction error between the sample counts
            and the reconstructed sample counts.

        - _n_parameters:
            The number of parameters fitted in CorrNMF

        - objective: str
            "minimize" or "maximize". Whether the NMF algorithm maximizes
            or minimizes the objective function. Some algorithms maximize a likelihood,
            others minimize a distance. The distinction is useful for filtering NMF
            runs based on the fitted objective function value.

        - corr_signatures: pd.DataFrame
            The signature correlation matrix induced by the signature embeddings

        - corr_samples: pd.DataFrame
            The sample correlation matrix induced by the sample embeddings


    The following methods are implemented in the abstract class CorrNMF:

        - objective_function:
            The evidence lower bound (ELBO) of the log-likelihood.
            Note: The ELBO is sometimes called the variational lower bound.

        - _surrogate_objective_function:
            A surrogate lower bound of the ELBO after introducing the
            auxiliary parameters p. In contrast to the original objective_function,
            the surrogate is strictly convex in the signature and sample embeddings

        - loglikelihood:
            The loglikelihood of the underyling generative model

        - _initialize:
            Initialize all model parameters and latent variables depending on the
            initialization method chosen

        - _get_embedding_annotations:
            A helper function to concatenate signature and sample names

        - plot_embeddings:
            Plot signature or sample embeddings in 2D using PCA, tSNE or UMAP.
            The respective plotting functions are implemented in the plot.py module

    More specific docstrings are written for the respective attributes and methods.
    """

    def __init__(
        self,
        n_signatures=1,
        dim_embeddings=None,
        init_method="nndsvd",
        min_iterations=500,
        max_iterations=10000,
        conv_test_freq=10,
        tol=1e-7,
    ):
        """
        Input:
        ------
        n_signatures: int
            The number of underlying signatures that are assumed to
            have generated the mutation count data

        dim_embeddings: int
            The assumed dimension of the signature and sample embeddings.
            Should be smaller or equal to the number of signatures as a dimension
            equal to the number of signatures covers the case of independent
            signatures. The smaller the embedding dimension, the stronger the
            enforced correlation structure  on both signatures and samples.

        init_method: str
            One of "custom", "flat", "hierarchical_cluster", "nndsvd",
            "nndsvda", "nndsvdar" "random" and "separableNMF".
            See the initialization module for further details.

        min_iterations: int
            The minimum number of iterations to perform during inference

        max_iterations: int
            The maximum number of iterations to perform during inference

        conv_test_freq: int
            The frequency at which the algorithm is tested for convergence.
            The objective function value is only computed every 'conv_test_freq'
            many iterations, which also affects a potentially saved history of
            the objective function values.

        tol: float
            The CorrNMF algorithm is converged when the relative change of the
            surrogate objective function of one iteration is smaller
            than the tolerance 'tol'.
        """
        super().__init__(
            n_signatures,
            init_method,
            min_iterations,
            max_iterations,
            conv_test_freq,
            tol,
        )

        if dim_embeddings is None:
            dim_embeddings = n_signatures

        self.dim_embeddings = dim_embeddings

        # initialize data/fitting-dependent attributes
        self.W = None
        self.alpha = None
        self.beta = None
        self.L = None
        self.U = None
        self.sigma_sq = None

    @property
    def signatures(self) -> pd.DataFrame:
        signatures = pd.DataFrame(
            self.W, index=self.mutation_types, columns=self.signature_names
        )
        return signatures

    @property
    def exposures(self) -> pd.DataFrame:
        """
        In contrast to the classical NMF framework, the exposure matrix is
        restructured and determined by the signature & sample biases and
        embeddings.
        """
        exposures = pd.DataFrame(
            np.exp(self.alpha + self.beta[:, np.newaxis] + self.L.T @ self.U),
            index=self.signature_names,
            columns=self.sample_names,
        )
        return exposures

    @property
    def _n_parameters(self):
        """
        There are n_features * n_signatures parameters corresponding to
        the signature matrix, each embedding corresponds to dim_embeddings parameters,
        and each signature & sample has a bias.
        Finally, the model variance is a single positive real number.

        Note: We do not include the number of auxiliary parameters p.
        """
        n_parameters_signatures = self.n_features * self.n_signatures
        n_parameters_embeddings = self.dim_embeddings * (
            self.n_signatures + self.n_samples
        )
        n_parameters_biases = self.n_samples + self.n_signatures
        n_parameters_exposures = n_parameters_embeddings + n_parameters_biases
        n_parameters = n_parameters_signatures + n_parameters_exposures + 1

        return n_parameters

    @property
    def reconstruction_error(self):
        return kl_divergence(self.X, self.W, self.exposures.values)

    @property
    def samplewise_reconstruction_error(self):
        return samplewise_kl_divergence(self.X, self.W, self.exposures.values)

    def objective_function(self, penalize_sample_embeddings=True) -> float:
        """
        The evidence lower bound (ELBO)
        """
        elbo = poisson_llh(self.X, self.signatures.values, self.exposures.values)
        elbo -= (
            0.5
            * self.dim_embeddings
            * self.n_signatures
            * np.log(2 * np.pi * self.sigma_sq)
        )
        elbo -= np.sum(self.L**2) / (2 * self.sigma_sq)

        if penalize_sample_embeddings:
            elbo -= (
                0.5
                * self.dim_embeddings
                * self.n_samples
                * np.log(2 * np.pi * self.sigma_sq)
            )
            elbo -= np.sum(self.U**2) / (2 * self.sigma_sq)

        return elbo

    @property
    def objective(self) -> str:
        return "maximize"

    def _surrogate_objective_function(self, penalize_sample_embeddings=True) -> float:
        """
        The surrogate lower bound of the ELBO.
        """
        p = self._update_p()
        exposures = self.exposures.values
        aux = np.log(self.W)[:, :, None] + np.log(exposures)[None, :, :] - np.log(p)
        sof_value = np.einsum("VD,VKD,VKD->", self.X, p, aux, optimize="greedy").item()
        sof_value -= np.sum(gammaln(1 + self.X))
        sof_value -= np.sum(exposures)
        sof_value -= (
            0.5
            * self.dim_embeddings
            * self.n_signatures
            * np.log(2 * np.pi * self.sigma_sq)
        )
        sof_value -= np.sum(self.L**2) / (2 * self.sigma_sq)

        if penalize_sample_embeddings:
            sof_value -= (
                0.5
                * self.dim_embeddings
                * self.n_samples
                * np.log(2 * np.pi * self.sigma_sq)
            )
            sof_value -= np.sum(self.U**2) / (2 * self.sigma_sq)

        return sof_value

    def loglikelihood(self):
        return self.objective_function()

    @abstractmethod
    def _update_alpha(self):
        pass

    @abstractmethod
    def _update_sigma_sq(self):
        pass

    @abstractmethod
    def _update_W(self):
        pass

    @abstractmethod
    def _update_p(self):
        pass

    @abstractmethod
    def _update_l(self, index, aux_row, outer_prods_U):
        r"""
        Input:
        ------
        index: int
            The index of the signature whose embedding is updated

        aux_row: nd.ndarray
            Row of the following matrix:
            aux_kd = \sum_v X_vd * p_vkd.
            This auxiliary matrix is used for updating the signatures
            and the sample embeddidngs. The aux_row argument
            is the k-th row of aux, where k is equal to 'index'.

        outer_prods_U: np.ndarray
            All outer products of the sample embeddings.
            shape: (n_samples, dim_embeddings, dim_embeddings)
        """

    @abstractmethod
    def _update_u(self, index, aux_col, outer_prods_L):
        r"""
        Input:
        ------
        index: int
            The index of the sample whose embedding is updated

        aux_col: nd.ndarray
            Column of the following matrix:
            aux_kd = \sum_v X_vd * p_vkd.
            This auxiliary matrix is used for updating the signatures
            and the sample embeddidngs. The aux_col argument
            is the d-th row of aux, where d is equal to 'index'.

        outer_prods_L: np.ndarray
            All outer products of the signature embeddings.
            shape: (n_signatures, dim_embeddings, dim_embeddings)
        """

    def _check_given_biases(self, given_biases, expected_n_biases, name):
        type_checker(name, given_biases, np.ndarray)
        shape_checker(name, given_biases, (expected_n_biases,))

    def _check_given_embeddings(self, given_embeddings, expected_n_embeddings, name):
        type_checker(name, given_embeddings, np.ndarray)
        shape_checker(
            name, given_embeddings, (self.dim_embeddings, expected_n_embeddings)
        )

    def _check_given_parameters(
        self,
        given_signatures,
        given_signature_biases,
        given_signature_embeddings,
        given_sample_biases,
        given_sample_embeddings,
    ):
        if given_signatures is not None:
            self._check_given_signatures(given_signatures)

        if given_signature_biases is not None:
            self._check_given_biases(
                given_signature_biases, self.n_signatures, "given_signature_biases"
            )

        if given_signature_embeddings is not None:
            self._check_given_embeddings(
                given_signature_embeddings,
                self.n_signatures,
                "given_signature_embeddings",
            )

        if given_sample_biases is not None:
            self._check_given_biases(
                given_sample_biases, self.n_samples, "given_sample_biases"
            )

        if given_sample_embeddings is not None:
            self._check_given_embeddings(
                given_sample_embeddings, self.n_samples, "given_sample_embeddings"
            )

    def _initialize(
        self,
        given_signatures=None,
        given_signature_biases=None,
        given_signature_embeddings=None,
        given_sample_biases=None,
        given_sample_embeddings=None,
        init_kwargs=None,
    ):
        """
        Initialize the signature matrix W, sample biases alpha, signature biases beta,
        the squared variance, and the signature and sample embeddings.
        The signatures or signature embeddings can also be provided by the user.

        Parameters
        ----------
        given_signatures: pd.DataFrame, default=None
            A priori known signatures. The number of given signatures has
            to be less or equal to the number of signatures of NMF
            algorithm instance, and the mutation type names have to match
            the mutation types of the count data.

        given_signature_biases : np.ndarray, default=None
            Known signature biases of shape (n_signatures,) that will be fixed
            during model fitting.

        given_signature_embeddings : np.ndarray, default=None
            A priori known signature embeddings of shape (dim_embeddings, n_signatures).

        given_sample_biases : np.ndarray, default=None
            Known sample biases of shape (n_samples,) that will be fixed
            during model fitting.

        given_sample_embeddings : np.ndarray, default=None
            A priori known sample embeddings of shape (dim_embeddings, n_samples).

        init_kwargs : dict
            Any further keyword arguments to pass to the initialization method.
            This includes, for example, a possible 'seed' keyword argument
            for all stochastic methods.
        """
        self._check_given_parameters(
            given_signatures,
            given_signature_biases,
            given_signature_embeddings,
            given_sample_biases,
            given_sample_embeddings,
        )

        if given_signatures is not None:
            self.n_given_signatures = len(given_signatures.columns)
        else:
            self.n_given_signatures = 0

        init_kwargs = {} if init_kwargs is None else init_kwargs.copy()
        self.W, _, self.signature_names = initialize(
            self.X, self.n_signatures, self.init_method, given_signatures, **init_kwargs
        )
        self.sigma_sq = 1.0

        if given_signature_biases is None:
            self.beta = np.zeros(self.n_signatures)
        else:
            self.beta = given_signature_biases

        if given_signature_embeddings is None:
            self.L = np.random.multivariate_normal(
                np.zeros(self.dim_embeddings),
                np.identity(self.dim_embeddings),
                size=self.n_signatures,
            ).T
        else:
            self.L = given_signature_embeddings

        if given_sample_biases is None:
            self.alpha = np.zeros(self.n_samples)
        else:
            self.alpha = given_sample_biases

        if given_sample_embeddings is None:
            self.U = np.random.multivariate_normal(
                np.zeros(self.dim_embeddings),
                np.identity(self.dim_embeddings),
                size=self.n_samples,
            ).T
        else:
            self.U = given_sample_embeddings

    @property
    def corr_signatures(self) -> pd.DataFrame:
        norms = np.sqrt(np.sum(self.L**2, axis=0))

        corr_vector = np.array(
            [
                np.dot(l1, l2) / (norms[k1] * norms[k1 + k2 + 1])
                for k1, l1 in enumerate(self.L.T)
                for k2, l2 in enumerate(self.L[:, (k1 + 1) :].T)
            ]
        )
        corr_matrix = squareform(corr_vector) + np.identity(self.n_signatures)
        corr = pd.DataFrame(
            corr_matrix, index=self.signature_names, columns=self.signature_names
        )

        return corr

    @property
    def corr_samples(self) -> pd.DataFrame:
        norms = np.sqrt(np.sum(self.U**2, axis=0))

        corr_vector = np.array(
            [
                np.dot(u1, u2) / (norms[d1] * norms[d1 + d2 + 1])
                for d1, u1 in enumerate(self.U.T)
                for d2, u2 in enumerate(self.U[:, (d1 + 1) :].T)
            ]
        )
        corr_matrix = squareform(corr_vector) + np.identity(self.n_samples)
        corr = pd.DataFrame(
            corr_matrix, index=self.sample_names, columns=self.sample_names
        )

        return corr

    def reorder(self, other_signatures, metric="cosine", keep_names=False):
        reordered_indices = match_signatures_pair(
            other_signatures, self.signatures, metric=metric
        )
        self.W = self.W[:, reordered_indices]
        self.beta = self.beta[reordered_indices]
        self.L = self.L[:, reordered_indices]

        if keep_names:
            self.signature_names = self.signature_names[reordered_indices]

        return reordered_indices

    def _get_embedding_data(self) -> np.ndarray:
        """
        In CorrNMF, the data for the embedding plot are the (transpoed) signature and
        sample embeddings.
        """
        return np.concatenate([self.L, self.U], axis=1).T.copy()

    def _get_default_embedding_annotations(self) -> np.ndarray:
        """
        The embedding plot defaults to annotating the signature embeddings.
        """
        return self.signature_names



from abc import ABC, abstractmethod

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from ..plot import (
    corr_plot,
    embeddings_plot,
    exposures_plot,
    history_plot,
    signatures_plot,
)
from ..utils import type_checker, value_checker

EPSILON = np.finfo(np.float32).eps


class SignatureNMF(ABC):
    """
    The abstract class SignatureNMF unifies the structure of
    multiple NMF algorithms used for signature analysis.

    Common properties and methods of all algorithms are indicated,
    i.e. have to be implemented by child classes, or implemented. Overview:

    Every child class has to implement the following attributes:

        - signatures: pd.DataFrame
            The signature matrix including mutation type names and signature names

        - exposures: pd.DataFrames
            The exposure matrix including the signature names and sample names

        - _n_parameters: int
            The number of parameters fitted by the NMF algorithm.
            This is needed to compute the Bayesian Information Criterion (BIC)

        - reconstruction_error: float
            The reconstruction error between the count matrix and
            the reconstructed count matrix.

        - samplewise_reconstruction_error: np.ndarray
            The samplewise reconstruction error between the sample counts and
            the reconstructed sample counts.

        - objective: str
            "minimize" or "maximize". Whether the NMF algorithm maximizes or
            minimizes the objective function. Some algorithms maximize a likelihood,
            others minimize a distance. The distinction is useful for filtering NMF runs
            based on the fitted objective function value downstream.

        - corr_signatures: pd.DataFrame
            The signature correlation matrix

        - corr_samples: pd.DataFrame
            The sample correlation matrix


    Every child class has to implement the following methods:

        - objective_fuction:
            The objective function to optimize when running the algorithm

        - loglikelihood:
            The loglikelihood of the underyling generative model

        - _initialize:
            A method to initialize all model parameters before fitting

        - fit:
            Run the NMF algorithm for a given mutation count data. Every
            fit method should also implement a version that allows fixing
            arbitrary many a priori known signatures.

        - plot_embeddings:
            Plot the sample (and potentially the signature) embeddings in 2D.


    The following attributes and methods are implemented in SignatureNMF:

        - data_reconstructed: pd.DataFrame
            The recovered mutation count data given
            the current signatures and exposures.

        - X_reconstructed: np.ndarray
            The recovered mutation count matrix given
            the current signatures and exposures

        - bic: float
            The value of the Bayesian Information Criterion (BIC)

        - _setup_parameters_fitting:
            Perform parameter checks and add the input mutation counts matrix
            as an attributes

        - plot_signatures:
            Plot the signatures using the signatures_plot function implemented in
            the plot module

        - plot_correlation:
            Plot the correlation of either the signatures or exposures
            using the corr_plot function implemented in the plot module

    More specific docstrings are written for the respective attributes and methods.
    """

    def __init__(
        self,
        n_signatures=1,
        init_method="nndsvd",
        min_iterations=500,
        max_iterations=10000,
        conv_test_freq=10,
        tol=1e-7,
    ):
        """
        Input:
        ------
        n_signatures: int
            The number of underlying signatures that are assumed to
            have generated the mutation count data

        init_method: str
            The initialization method for the NMF algorithm

        min_iterations: int
            The minimum number of iterations to perform by the NMF algorithm

        max_iterations: int
            The maximum number of iterations to perform by the NMF algorithm

        conv_test_freq: int
            The frequency at which the algorithm is tested for convergence.
            The objective function value is only computed every 'conv_test_freq'
            many iterations, which also affects a potentially saved history of
            the objective function values.

        tol: float
            The NMF algorithm is converged when the relative change of
            the objective function of one iteration is smaller
            than the tolerance 'tol'.
        """
        init_methods = [
            "custom",
            "flat",
            "hierarchical_cluster",
            "nndsvd",
            "nndsvda",
            "nndsvdar",
            "random",
            "separableNMF",
        ]
        value_checker("init_method", init_method, init_methods)

        self.n_signatures = n_signatures
        self.signature_names = None
        self.init_method = init_method
        self.min_iterations = min_iterations
        self.max_iterations = max_iterations
        self.conv_test_freq = conv_test_freq
        self.tol = tol

        # initialize data/fitting dependent attributes
        self.X = None
        self.n_features = 0
        self.n_given_signatures = 0
        self.n_samples = 0
        self.mutation_types = np.empty(0, dtype=str)
        self.sample_names = np.empty(0, dtype=str)
        self.history = {}

    @property
    @abstractmethod
    def signatures(self) -> pd.DataFrame:
        """
        Extract the mutational signatures as a pandas dataframe.
        """
        pass

    @property
    @abstractmethod
    def exposures(self) -> pd.DataFrame:
        """
        Extract the signature exposures of samples as a pandas dataframe.
        """
        pass

    @property
    def data_reconstructed(self) -> pd.DataFrame:
        return (self.signatures @ self.exposures).astype(int)

    @property
    def X_reconstructed(self) -> np.ndarray:
        return self.data_reconstructed.values

    @property
    @abstractmethod
    def reconstruction_error(self) -> float:
        """
        The reconstruction error between the count matrix and
        the reconstructed count matrix.
        """
        pass

    @property
    @abstractmethod
    def samplewise_reconstruction_error(self) -> np.ndarray:
        """
        The samplewise reconstruction error between the sample counts and
        the reconstructed sample counts.
        """
        pass

    @abstractmethod
    def objective_function(self) -> float:
        """
        The objective function to be optimized during fitting.
        """
        pass

    @abstractmethod
    def loglikelihood(self) -> float:
        """
        The log-likelihood of the underlying generative model.
        """
        pass

    @property
    @abstractmethod
    def _n_parameters(self) -> int:
        """
        Every child class has to implement a function returning
        the number of parameters estimated by the respective model.
        This is allows to, for example, implement the BIC
        (Bayesian information criterion). The BIC can be used to
        estimate the optimal number of signatures.
        """
        pass

    @property
    def bic(self) -> float:
        """
        Bayesian information criterion (BIC).
        Can only be called after the _setup_parameters_fitting function as it
        requires the number of samples be an attribute.
        """
        return self._n_parameters * np.log(self.n_samples) - 2 * self.loglikelihood()

    def _check_given_signatures(self, given_signatures: pd.DataFrame):
        """
        Check if the given signatures are compatible with the
        number of signatures of the algorithm and the
        mutation types of the input data.

        given_signatures: pd.DataFrame
            Known signatures that should be fixed by the algorithm.
            The number of known signatures can be less or equal to the
            number of signatures specified in the algorithm instance.
        """
        type_checker("given_signatures", given_signatures, pd.DataFrame)
        given_mutation_types = given_signatures.index.to_numpy(dtype=str)
        compatible = (
            np.array_equal(given_mutation_types, self.mutation_types)
            and given_signatures.shape[1] <= self.n_signatures
        )

        if not compatible:
            raise ValueError(
                f"You have to provide at most {self.n_signatures} signatures with "
                f"mutation types matching to your data."
            )

    @abstractmethod
    def _initialize(self):
        """
        Initialize model parameters and attributes before fitting.
        Enforcing the existence of _initialize unifies the implementation of
        the NMF algorithms.

        Example:

            Before running the Lee & Seung NMF multiplicative update rules to
            decompose the mutation count matrix X into a signature matrix W and
            an exposure matrix H, both W and H have to be initialized.
        """

    def _setup_data_parameters(self, data: pd.DataFrame):
        """
        Perform parameter checks before running the fit method.

        Input:
        ------
        data: pd.DataFrame
            The mutation count pandas dataframe with indices and column names.
            Samples are expected to corresponding to columns.
        """
        type_checker("data", data, pd.DataFrame)
        self.X = data.values.clip(EPSILON)
        self.n_features, self.n_samples = data.shape
        self.mutation_types = data.index.values.astype(str)
        self.sample_names = data.columns.values.astype(str)

    @abstractmethod
    def fit(self, data: pd.DataFrame, given_signatures=None):
        """
        Fit the model parameters. Child classes are expected to handle
        'given_signatures' appropriately.

        Input:
        ------
        data: pd.DataFrame
            The named mutation count data of shape (n_features, n_samples).

        given_signatures: pd.DataFrame, by default None
            A priori known signatures. The number of given signatures has
            to be less or equal to the number of signatures of NMF
            algorithm instance, and the mutation type names have to match
            the mutation types of the count data.
        """

    def plot_history(self, ax=None, min_iteration=0, outfile=None, **kwargs):
        if not self.history:
            raise ValueError(
                "No history available, the model has to be fitted first. "
                "Remember to set 'history' to 'True' when calling 'fit()'."
            )

        history_plot(
            self.history["objective_function"],
            self.conv_test_freq,
            min_iteration=min_iteration,
            ax=ax,
            **kwargs,
        )

        if outfile is not None:
            plt.savefig(outfile, bbox_inches="tight")

        return ax

    def plot_signatures(
        self,
        catalog=None,
        colors=None,
        annotate_mutation_types=False,
        axes=None,
        outfile=None,
        **kwargs,
    ):
        """
        Plot the signatures, see plot.py for the implementation of signatures_plot.
        """
        axes = signatures_plot(
            self.signatures,
            catalog=catalog,
            colors=colors,
            annotate_mutation_types=annotate_mutation_types,
            axes=axes,
            **kwargs,
        )

        if outfile is not None:
            plt.savefig(outfile, bbox_inches="tight")

        return axes

    def plot_exposures(
        self,
        sample_order=None,
        reorder_signatures=True,
        annotate_samples=True,
        colors=None,
        ncol_legend=1,
        ax=None,
        outfile=None,
        **kwargs,
    ):
        """
        Visualize the exposures as a stacked bar chart,
        see plot.py for the implementation.
        """
        ax = exposures_plot(
            exposures=self.exposures,
            sample_order=sample_order,
            reorder_signatures=reorder_signatures,
            annotate_samples=annotate_samples,
            colors=colors,
            ncol_legend=ncol_legend,
            ax=ax,
            **kwargs,
        )
        if outfile is not None:
            plt.savefig(outfile, bbox_inches="tight")

        return ax

    @property
    @abstractmethod
    def corr_signatures(self) -> pd.DataFrame:
        """
        Every child class of SignatureNMF has to implement a function that
        returns the signature correlation matrix as a pandas dataframe.
        """

    @property
    @abstractmethod
    def corr_samples(self) -> pd.DataFrame:
        """
        Every child class of SignatureNMF has to implement a function that
        returns the sample correlation matrix as a pandas dataframe.
        """

    def plot_correlation(self, data="signatures", annot=False, outfile=None, **kwargs):
        """
        Plot the correlation matrix of the signatures or samples.
        See plot.py for the implementation of corr_plot.

        Input:
        ------
        *args, **kwargs:
            arguments to be passed to corr_plot
        """
        value_checker("data", data, ["signatures", "samples"])

        if data == "signatures":
            corr = self.corr_signatures

        else:
            corr = self.corr_samples

        clustergrid = corr_plot(corr, annot=annot, **kwargs)

        if outfile is not None:
            plt.savefig(outfile, bbox_inches="tight")

        return clustergrid

    @abstractmethod
    def _get_embedding_data(self) -> np.ndarray:
        """
        Get the data points for the dimensionality reduction / embedding plot.
        One data point corresponds to a row of the embedding data.
        Usually, these are the transposed exposures.
        """

    @abstractmethod
    def _get_default_embedding_annotations(self) -> np.ndarray:
        """
        Get the annotations of the data points in the embedding plot.
        """

    def plot_embeddings(self, annotations=None, outfile=None, **kwargs):
        """
        Plot a dimensionality reduction of the exposure representation.
        In most NMF algorithms, this is just the exposures of the samples.
        In CorrNMF, the exposures matrix is refactored, and there are both
        sample and signature embeddings in a shared embedding space.

        If the embedding dimension is one or two, the embeddings are be plotted
        directly, ignoring the chosen method.
        See plot.py for the implementation of 'embeddings_plot'.

        Parameters
        ----------
        annotations : list[str], default=None
            Annotations per data point, e.g. the sample names. If None,
            the algorithm-specific default annotations are used.
            For example, CorrNMF annotates the signature embeddings by default.
            Note that there are 'n_signatures' + 'n_samples' data points in CorrNMF,
            i.e. the first 'n_signatures' elements in 'annotations'
            are the signature annotations, not any sample annotations.

        outfile : str, default=None
            If not None, the figure will be saved in the specified file path.

        **kwargs :
            keyword arguments to pass to seaborn's scatterplot

        Returns
        -------
        ax : matplotlib.axes.Axes
            The matplotlib axes containing the plot.
        """
        # one data point corresponds to a row of embedding_data
        embedding_data = self._get_embedding_data()

        if annotations is None:
            annotations = self._get_default_embedding_annotations()

        ax = embeddings_plot(data=embedding_data, annotations=annotations, **kwargs)

        if outfile is not None:
            plt.savefig(outfile, bbox_inches="tight")

        return ax



"""
Multimodal correlated NMF (MultiCorrNMF) fits multiple correlated NMF (CorrNMF)
models jointly in the following manner:
Assuming that the input data for each modality originates from the identical samples,
MultiCorrNMF fixes the sample embeddings accross modalities and learns signature
embeddings for all modalities in the same embedding space.
"""
# This implementation heavily relies on the implementaion of CorrNMF in
# corrnmf_det.py. In particular, CorrNMFDet methods with a leading '_'
# are accessed.
# pylint: disable=protected-access

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy import optimize
from scipy.spatial.distance import squareform

from ..plot import (
    _get_sample_order,
    corr_plot,
    embeddings_plot,
    history_plot,
    signatures_plot,
)
from ..utils import type_checker, value_checker
from . import _utils_corrnmf
from .corrnmf_det import CorrNMFDet

EPSILON = np.finfo(np.float32).eps


class MultimodalCorrNMF:
    def __init__(
        self,
        n_modalities,
        ns_signatures=None,
        dim_embeddings=None,
        init_method="nndsvd",
        min_iterations=500,
        max_iterations=10000,
        conv_test_freq=10,
        tol=1e-7,
    ):
        self.n_modalities = n_modalities

        if ns_signatures is None:
            ns_signatures = np.ones(n_modalities, dtype=int)

        self.ns_signatures = ns_signatures

        if dim_embeddings is None:
            dim_embeddings = np.max(ns_signatures)

        self.dim_embeddings = dim_embeddings
        self.init_method = init_method
        self.min_iterations = min_iterations
        self.max_iterations = max_iterations
        self.conv_test_freq = conv_test_freq
        self.tol = tol
        self.models = [
            CorrNMFDet(n_signatures, dim_embeddings, init_method)
            for n_signatures in ns_signatures
        ]

        # initialize data/fitting dependent attributes
        self.modality_names = np.empty(n_modalities, dtype=str)
        self.n_samples = 0
        self.history = {}

    @property
    def signatures(self) -> dict:
        return {
            name: model.signatures
            for name, model in zip(self.modality_names, self.models)
        }

    @property
    def exposures(self) -> dict:
        return {
            name: model.exposures
            for name, model in zip(self.modality_names, self.models)
        }

    @property
    def data_reconstructed(self) -> dict:
        return {
            name: model.data_reconstructred
            for name, model in zip(self.modality_names, self.models)
        }

    @property
    def Xs_reconstructed(self) -> np.ndarray:
        return {
            name: model.X_reconstructed
            for name, model in zip(self.modality_names, self.models)
        }

    @property
    def reconstruction_errors(self) -> float:
        return {
            name: model.reconstruction_error
            for name, model in zip(self.modality_names, self.models)
        }

    @property
    def samplewise_reconstruction_errors(self) -> np.ndarray:
        return {
            name: model.samplewise_reconstruction_error
            for name, model in zip(self.modality_names, self.models)
        }

    def objective_function(self) -> float:
        """
        The objective function to be optimized during fitting.
        """
        elbo = np.sum(
            [
                model.objective_function(penalize_sample_embeddings=False)
                for model in self.models
            ]
        )
        elbo -= (
            0.5
            * self.dim_embeddings
            * self.n_samples
            * np.log(2 * np.pi * self.models[0].sigma_sq)
        )
        elbo -= np.sum(self.models[0].U ** 2) / (2 * self.models[0].sigma_sq)
        return elbo

    @property
    def objective(self) -> str:
        return "maximize"

    def _surrogate_objective_function(self) -> float:
        """
        The surrogate lower bound of the ELBO.
        """
        ps = self._update_ps()
        sof_value = np.sum(
            [
                model._surrogate_objective_function(penalize_sample_embeddings=False)
                for model, p in zip(self.models, ps)
            ]
        )
        sof_value -= (
            0.5
            * self.dim_embeddings
            * self.n_samples
            * np.log(2 * np.pi * self.models[0].sigma_sq)
        )
        sof_value -= np.sum(self.models[0].U ** 2) / (2 * self.models[0].sigma_sq)
        return sof_value

    def loglikelihood(self) -> float:
        """
        The log-likelihood of the underlying generative model.
        """
        return self.objective_function()

    @property
    def _n_parameters(self) -> int:
        n_parameters_signatures = np.sum(
            [model.n_features * model.n_signatures for model in self.models]
        )
        n_parameters_embeddings = self.dim_embeddings * (
            np.sum(self.ns_signatures) + self.n_samples
        )
        n_parameters_biases = self.n_modalities * self.n_samples + np.sum(
            self.ns_signatures
        )
        n_parameters_exposures = n_parameters_embeddings + n_parameters_biases
        n_parameters = n_parameters_signatures + n_parameters_exposures + 1

        return n_parameters

    @property
    def bic(self) -> float:
        return self._n_parameters * np.log(self.n_samples) - 2 * self.loglikelihood()

    def _update_alphas(self, given_sample_biases):
        for model, given_sam_biases in zip(self.models, given_sample_biases):
            model._update_alpha(given_sam_biases)

    def _update_betas(self, ps, given_signature_biases):
        for model, p, given_sig_biases in zip(self.models, ps, given_signature_biases):
            model._update_beta(p, given_sig_biases)

    def _update_sigma_sq(self):
        Ls = np.concatenate([model.L for model in self.models], axis=1)
        embeddings = np.concatenate([Ls, self.models[0].U], axis=1)
        sigma_sq = np.mean(embeddings**2)
        sigma_sq = np.clip(sigma_sq, EPSILON, None)

        for model in self.models:
            model.sigma_sq = sigma_sq

    def _update_Ws(self):
        for model in self.models:
            if model.n_given_signatures < model.n_signatures:
                model._update_W()

    def _update_ps(self):
        return [model._update_p() for model in self.models]

    def _objective_fun_u(self, u, index, aux_cols):
        sigma_sq = self.models[0].sigma_sq
        s = -np.sum(
            [
                _utils_corrnmf.objective_function_embedding(
                    u,
                    model.L,
                    model.alpha[index],
                    model.beta,
                    sigma_sq,
                    aux_col,
                    add_penalty=False,
                )
                for model, aux_col in zip(self.models, aux_cols)
            ]
        )
        s -= np.dot(u, u) / (2 * sigma_sq)
        return -s

    def _gradient_u(self, u, index, s_grads):
        sigma_sq = self.models[0].sigma_sq
        s = -np.sum(
            [
                _utils_corrnmf.gradient_embedding(
                    u,
                    model.L,
                    model.alpha[index],
                    model.beta,
                    sigma_sq,
                    s_grad,
                    add_penalty=False,
                )
                for model, s_grad in zip(self.models, s_grads)
            ],
            axis=0,
        )
        s -= u / sigma_sq
        return -s

    def _hessian_u(self, u, index, outer_prods_Ls):
        sigma_sq = self.models[0].sigma_sq
        s = -np.sum(
            [
                _utils_corrnmf.hessian_embedding(
                    u,
                    model.L,
                    model.alpha[index],
                    model.beta,
                    sigma_sq,
                    outer_prods_L,
                    add_penalty=False,
                )
                for model, outer_prods_L in zip(self.models, outer_prods_Ls)
            ],
            axis=0,
        )
        s -= np.diag(np.full(self.dim_embeddings, 1 / sigma_sq))
        return -s

    def _update_u(self, index, aux_cols, outer_prods_Ls):
        def objective_fun(u):
            return self._objective_fun_u(u, index, aux_cols)

        s_grads = np.array(
            [
                np.sum(aux_col * model.L, axis=1)
                for model, aux_col in zip(self.models, aux_cols)
            ]
        )

        def gradient(u):
            return self._gradient_u(u, index, s_grads)

        def hessian(u):
            return self._hessian_u(u, index, outer_prods_Ls)

        u = optimize.minimize(
            fun=objective_fun,
            x0=self.models[0].U[:, index],
            method="Newton-CG",
            jac=gradient,
            hess=hessian,
            options={"maxiter": 3},
        ).x
        u[(0 < u) & (u < EPSILON)] = EPSILON
        u[(-EPSILON < u) & (u < 0)] = -EPSILON

        for model in self.models:
            model.U[:, index] = u

    def _update_U(self, auxs):
        outer_prods_Ls = [
            np.einsum("mK,nK->Kmn", model.L, model.L) for model in self.models
        ]

        for d in range(self.n_samples):
            aux_cols = [aux[:, d] for aux in auxs]
            self._update_u(d, aux_cols, outer_prods_Ls)

    def _update_Ls(self, auxs, outer_prods_U, given_signature_embeddings):
        for model, aux, given_sig_embs in zip(
            self.models, auxs, given_signature_embeddings
        ):
            if given_sig_embs is None:
                model._update_L(aux, outer_prods_U)

    def _update_LsU(self, ps, given_signature_embeddings, given_sample_embeddings):
        auxs = [
            np.einsum("vd,vkd->kd", model.X, p) for model, p in zip(self.models, ps)
        ]
        outer_prods_U = np.einsum("mD,nD->Dmn", self.models[0].U, self.models[0].U)
        self._update_Ls(auxs, outer_prods_U, given_signature_embeddings)

        if given_sample_embeddings is None:
            self._update_U(auxs)

    def _setup_data_parameters(self, data: list):
        type_checker("data", data, list)

        if len(data) != self.n_modalities:
            raise ValueError(
                f"The input data has to be {self.n_modalities} "
                "many named pandas dataframes."
            )

        for df in data:
            type_checker("input dataframe", df, pd.DataFrame)

            if df.index.name is None:
                raise ValueError(
                    "You have to set 'df.index.name' to a "
                    "meaningful name for every input dataframe."
                )

        self.modality_names = np.array([df.index.name for df in data])
        self.n_samples = data[0].shape[1]

        for model, df in zip(self.models, data):
            model._setup_data_parameters(df)

    def _initialize(
        self,
        given_signatures=None,
        given_signature_biases=None,
        given_signature_embeddings=None,
        given_sample_biases=None,
        given_sample_embeddings=None,
        init_kwargs=None,
    ):
        if given_signatures is None:
            given_signatures = [None for _ in range(self.n_modalities)]

        if given_signature_biases is None:
            given_signature_biases = [None for _ in range(self.n_modalities)]

        if given_signature_embeddings is None:
            given_signature_embeddings = [None for _ in range(self.n_modalities)]

        if given_sample_biases is None:
            given_sample_biases = [None for _ in range(self.n_modalities)]

        if given_sample_embeddings is None:
            U = np.random.multivariate_normal(
                np.zeros(self.dim_embeddings),
                np.identity(self.dim_embeddings),
                size=self.n_samples,
            ).T
        else:
            U = given_sample_embeddings

        for (
            model,
            modality_name,
            given_sigs,
            given_sig_biases,
            given_sig_embs,
            given_sam_biases,
        ) in zip(
            self.models,
            self.modality_names,
            given_signatures,
            given_signature_biases,
            given_signature_embeddings,
            given_sample_biases,
        ):
            model._initialize(
                given_signatures=given_sigs,
                given_signature_biases=given_sig_biases,
                given_signature_embeddings=given_sig_embs,
                given_sample_biases=given_sam_biases,
                given_sample_embeddings=U,
                init_kwargs=init_kwargs,
            )
            model.signature_names = np.concatenate(
                [
                    model.signature_names[: model.n_given_signatures],
                    np.char.add(
                        modality_name + " ",
                        model.signature_names[model.n_given_signatures :],
                    ),
                ]
            )
        return (
            given_signature_biases,
            given_signature_embeddings,
            given_sample_biases,
            given_sample_embeddings,
        )

    def fit(
        self,
        data: list,
        given_signatures=None,
        given_signature_biases=None,
        given_signature_embeddings=None,
        given_sample_biases=None,
        given_sample_embeddings=None,
        init_kwargs=None,
        history=False,
        verbose=0,
    ):
        self._setup_data_parameters(data)
        (
            given_signature_biases,
            given_signature_embeddings,
            given_sample_biases,
            given_sample_embeddings,
        ) = self._initialize(
            given_signatures=given_signatures,
            given_signature_biases=given_signature_biases,
            given_signature_embeddings=given_signature_embeddings,
            given_sample_biases=given_sample_biases,
            given_sample_embeddings=given_sample_embeddings,
            init_kwargs=init_kwargs,
        )
        of_values = [self.objective_function()]
        n_iteration = 0
        converged = False

        while not converged:
            n_iteration += 1

            if verbose and n_iteration % 100 == 0:
                print(f"iteration: {n_iteration}; objective: {of_values[-1]:.2f}")

            self._update_alphas(given_sample_biases)
            ps = self._update_ps()
            self._update_betas(ps, given_signature_biases)
            self._update_LsU(ps, given_signature_embeddings, given_sample_embeddings)
            self._update_sigma_sq()
            self._update_Ws()

            if n_iteration % self.conv_test_freq == 0:
                prev_of_value = of_values[-1]
                of_values.append(self.objective_function())
                rel_change = (of_values[-1] - prev_of_value) / np.abs(prev_of_value)
                converged = rel_change < self.tol and n_iteration >= self.min_iterations

            converged |= n_iteration >= self.max_iterations

        if history:
            self.history["objective_function"] = of_values[1:]

        return self

    def plot_history(self, ax=None, min_iteration=0, outfile=None, **kwargs):
        if not self.history:
            raise ValueError(
                "No history available, the model has to be fitted first. "
                "Remember to set 'history' to 'True' when calling 'fit()'."
            )

        history_plot(
            self.history["objective_function"],
            self.conv_test_freq,
            min_iteration=min_iteration,
            ax=ax,
            **kwargs,
        )

        if outfile is not None:
            plt.savefig(outfile, bbox_inches="tight")

        return ax

    def plot_signatures(
        self,
        colors=None,
        annotate_mutation_types=False,
        figsize=None,
        outfile=None,
        **kwargs,
    ):
        if colors is None:
            colors = [None for _ in range(self.n_modalities)]

        max_n_signatures = np.max(self.ns_signatures)

        if figsize is None:
            figsize = (8 * self.n_modalities, 2 * max_n_signatures)

        fig, axes = plt.subplots(max_n_signatures, self.n_modalities, figsize=figsize)

        for axs, model, cols in zip(axes.T, self.models, colors):
            model.plot_signatures(
                colors=cols,
                annotate_mutation_types=annotate_mutation_types,
                axes=axs[: model.n_signatures],
                **kwargs,
            )

            for ax in axs[model.n_signatures :]:
                fig.delaxes(ax)

        plt.tight_layout()

        if outfile is not None:
            plt.savefig(outfile, bbox_inches="tight")

        return axes

    def plot_exposures(
        self,
        sample_order=None,
        reorder_signatures=True,
        annotate_samples=True,
        colors=None,
        ncol_legend=1,
        axes=None,
        outfile=None,
        **kwargs,
    ):
        """
        Visualize the exposures as a stacked bar chart,
        see plot.py for the implementation.

        Input:
        ------
        **kwargs:
            arguments to be passed to exposure_plot
        """
        if axes is None:
            _, axes = plt.subplots(
                self.n_modalities, figsize=(20, 3 * self.n_modalities)
            )

        if colors is None:
            colors = [None for _ in range(self.n_modalities)]

        if sample_order is None:
            all_exposures = [model.exposures for model in self.models]
            all_exposures_normalized = pd.concat(
                [df / df.sum(axis=0) for df in all_exposures]
            )
            sample_order = _get_sample_order(all_exposures_normalized)

        for n, (ax, model, cols) in enumerate(zip(axes, self.models, colors)):
            if n < self.n_modalities - 1:
                annotate = False
            else:
                annotate = annotate_samples

            ax = model.plot_exposures(
                sample_order=sample_order,
                reorder_signatures=reorder_signatures,
                annotate_samples=annotate,
                colors=cols,
                ncol_legend=ncol_legend,
                ax=ax,
                **kwargs,
            )
            ax.set_title(f"{self.modality_names[n]} signature exposures")

        plt.tight_layout()

        if outfile is not None:
            plt.savefig(outfile, bbox_inches="tight")

        return axes

    @property
    def corr_signatures(self) -> pd.DataFrame:
        Ls = np.concatenate([model.L for model in self.models], axis=1)
        signature_names = np.concatenate(
            [model.signature_names for model in self.models]
        )
        norms = np.sqrt(np.sum(Ls**2, axis=0))

        corr_vector = np.array(
            [
                np.dot(l1, l2) / (norms[k1] * norms[k1 + k2 + 1])
                for k1, l1 in enumerate(Ls.T)
                for k2, l2 in enumerate(Ls[:, k1 + 1 :].T)
            ]
        )
        corr_matrix = squareform(corr_vector) + np.identity(np.sum(self.ns_signatures))
        corr = pd.DataFrame(corr_matrix, index=signature_names, columns=signature_names)

        return corr

    @property
    def corr_samples(self) -> pd.DataFrame:
        return self.models[0].corr_samples

    def plot_correlation(self, data="signatures", annot=False, outfile=None, **kwargs):
        """
        Plot the correlation matrix of the signatures or samples.
        See plot.py for the implementation of corr_plot.

        Input:
        ------
        *args, **kwargs:
            arguments to be passed to corr_plot
        """
        value_checker("data", data, ["signatures", "samples"])

        if data == "signatures":
            corr = self.corr_signatures

        else:
            corr = self.corr_samples

        clustergrid = corr_plot(corr, annot=annot, **kwargs)

        if outfile is not None:
            plt.savefig(outfile, bbox_inches="tight")

        return clustergrid

    def plot_embeddings(self, annotations=None, outfile=None, **kwargs):
        """
        Plot the signature and sample embeddings. If the embedding dimension
        is two, the embeddings will be plotted directly, ignoring the chosen method.
        See plot.py for the implementation of 'embeddings_plot'.

        Parameters
        ----------
        annotations : list[str], default=None
            Annotations per data point, e.g. the sample names. If None,
            all signatures are annotated.
            Note that there are sum('ns_signatures') + 'n_samples' data points,
            i.e. the first sum('ns_signatures') elements in 'annotations'
            are the signature annotations, not any sample annotations.

        outfile : str, default=None
            If not None, the figure will be saved in the specified file path.

        **kwargs :
            keyword arguments to pass to seaborn's scatterplot

        Returns
        -------
        ax : matplotlib.axes.Axes
            The matplotlib axes containing the plot.
        """
        Ls = np.concatenate([model.L for model in self.models], axis=1)
        embedding_data = np.concatenate([Ls, self.models[0].U], axis=1).T.copy()

        if annotations is None:
            annotations = np.concatenate(
                [model.signature_names for model in self.models]
            )

        ax = embeddings_plot(data=embedding_data, annotations=annotations, **kwargs)

        if outfile is not None:
            plt.savefig(outfile, bbox_inches="tight")

        return ax

    def feature_change(self, in_modality=None, out_modalities="all", normalize=True):
        if in_modality is None:
            in_modality = self.modality_names[0]

        in_model = self.models[list(self.modality_names).index(in_modality)]

        if out_modalities == "all":
            out_modalities = self.modality_names

        if type(out_modalities) is str:
            out_modalities = [out_modalities]

        out_modalities = [name for name in out_modalities if name != in_modality]
        out_modalities_indices = [
            n for n, name in enumerate(self.modality_names) if name in out_modalities
        ]
        results = [in_model.signatures]

        for n in out_modalities_indices:
            result = self.models[n].signatures @ np.exp(self.models[n].L.T @ in_model.L)
            result.columns = in_model.signature_names

            if normalize:
                result = result / result.sum(axis=0)

            results.append(result)

        return results

    def plot_feature_change(
        self,
        in_modality=None,
        out_modalities="all",
        colors=None,
        annotate_mutation_types=False,
        figsize=None,
        outfile=None,
        **kwargs,
    ):
        """
        For the signatures of one modality, plot the co-occuring spectra
        in other modalities. This is achieved by interpreting a signature
        embedding as a sample embedding and using the resulting exposures to
        compute distributions over mutation types in different modalities.

        Parameters
        ----------
        in_modality : str
            The modality name of the signatures of interest, e.g. "SBS".

        out_modalities : list or str, default="all"
            A list of modalities to convert the 'in_modality' signatures
            into, e.g. ["Indel", "SV"]. A single string can also be provided
            to select only one 'out_modality'. By default, all modalities
            other than the 'in_modality' are selected.

        colors : list, default=None
            A list of length '1 + len(out_modalities)' of colors to use
            for the signature plots of the input modalitiy signatures
            and the co-occuring spectra in the output modalites.

        annotate_mutation_types : bool, default=False
            If True, the x-axis of the spectra plots will be annotated
            with the mutation types of the respective modalities.

        figsize : tuple, default=None
            The size of the matplotlib figure. If None, the figure size
            is computed internally based on the number of input
            signatures and output modalities.

        outfile : str, default=None
            If not None, the figure will be saved to the provided path.

        kwargs : dict
            Any keyword arguments to be passed to matplotlibs ax.bar.

        Returns
        -------
        axes : np.ndarray
            An array of matplotlib axes containing the plots.
        """
        # result[0] are the 'in_modality' signatures
        results = self.feature_change(in_modality, out_modalities)
        n_signatures = results[0].shape[1]
        n_feature_spaces = len(results)

        if colors is None:
            colors = [None for _ in range(n_feature_spaces)]

        if figsize is None:
            figsize = (4 * n_feature_spaces, n_signatures)

        fig, axes = plt.subplots(n_signatures, n_feature_spaces, figsize=figsize)
        fig.suptitle("Signature feature change")

        for axs, result, cols in zip(axes.T, results, colors):
            signatures_plot(
                result,
                colors=cols,
                annotate_mutation_types=annotate_mutation_types,
                axes=axs,
                **kwargs,
            )

        plt.tight_layout()

        if outfile is not None:
            plt.savefig(outfile, bbox_inches="tight")

        return axes



import numpy as np
from numba import njit

EPSILON = np.finfo(np.float32).eps


@njit
def update_alpha(
    X: np.ndarray, beta: np.ndarray, L: np.ndarray, U: np.ndarray
) -> np.ndarray:
    """
    Compute the new sample biases alpha according to the update rule of CorrNMF.

    Parameters
    ----------
    X : np.ndarray of shape (n_features, n_samples)
        data matrix

    beta : np.ndarray of shape (n_signatures,)
        signature biases

    L : np.ndarray of shape (dim_embeddings, n_signatures)
        signature embeddings

    U : np.ndarray of shape (dim_embeddings, n_samples)
        sample embeddings

    Returns
    -------
    alpha : np.ndarray of shape (n_samples,)
        The new sample biases
    """
    second_term = np.exp(beta[:, np.newaxis] + L.T @ U)
    alpha = np.log(np.sum(X, axis=0)) - np.log(np.sum(second_term, axis=0))
    return alpha


@njit
def update_beta(
    X: np.ndarray, p: np.ndarray, alpha: np.ndarray, L: np.ndarray, U: np.ndarray
) -> np.ndarray:
    """
    Compute the new signature biases beta according to the update rule of CorrNMF.

    Parameters
    ----------
    X : np.ndarray of shape (n_features, n_samples)
        data matrix

    alpha : np.ndarray of shape (n_samples,)
        sample biases

    L : np.ndarray of shape (dim_embeddings, n_signatures)
        signature embeddings

    U : np.ndarray of shape (dim_embeddings, n_samples)
        sample embeddings

    Returns
    -------
    beta : np.ndarray of shape (n_signatures,)
        The new signature biases
    """
    n_features, n_signatures, n_samples = p.shape
    first_term = np.zeros(n_signatures)

    for k in range(n_signatures):
        for v in range(n_features):
            for d in range(n_samples):
                first_term[k] += X[v, d] * p[v, k, d]

    second_term = np.exp(alpha + L.T @ U)
    beta = np.log(first_term) - np.log(np.sum(second_term, axis=1))
    return beta


@njit
def update_p_unnormalized(W: np.ndarray, H: np.ndarray) -> np.ndarray:
    """
    Compute the new auxiliary parameters according to the update rule of CorrNMF.
    The normalization per mutation type and sample is not performed yet.

    Parameters
    ----------
    W : np.ndarray of shape (n_features, n_signatures)
        signature matrix

    H : np.ndarray of shape (n_signatures, n_samples)
        exposure matrix

    Returns
    -------
    p: np.ndarray of shape (n_features, n_signatures, n_samples)
    """
    n_features, n_signatures = W.shape
    n_samples = H.shape[1]
    p = np.zeros((n_features, n_signatures, n_samples))

    for v in range(n_features):
        for k in range(n_signatures):
            for d in range(n_samples):
                p[v, k, d] = W[v, k] * H[k, d]

    return p


@njit
def objective_function_embedding(
    embedding,
    embeddings_other,
    alpha,
    beta,
    sigma_sq,
    aux_vec,
    add_penalty=True,
):
    r"""
    The negative objective function of a signature or sample embedding in CorrNMF.

    Parameters
    ----------
    embedding : np.ndarray of shape (dim_embeddings,)
        The signature or sample embedding

    embeddings_other : np.ndarray of shape (dim_embeddings, n_samples | n_signatures)
        If 'embedding' is a signature embedding, 'embeddings_other' are
        all sample embeddings. If 'embedding' is a sample embedding,
        'embeddings_other' are all signature embeddings.

    alpha : float | np.narray of shape (n_samples,)
        If 'embedding' is a signature embedding, 'alpha' are
        all sample biases. If 'embedding' is a sample embedding,
        'alpha' is the bias of the corresponding sample.

    beta : np.array of shape (n_signatures,) | float
        If 'embedding' is a signature embedding, 'beta' is
        the corresponding signature bias. If 'embedding' is a sample embedding,
        'beta' are all signature biases.

    sigma_sq : float
        model variance

    aux_vec : np.ndarray of shape (n_signatures | n_samples,)
        A row or column of
        aux[k, d] = \sum_v X_vd * p_vkd,
        where X is the data matrix and p are the auxiliary parameters of CorrNMF.
        If 'embedding' is a signature embedding, the corresponding row is provided.
        If 'embedding' is a sample embedding, the corresponding column is provided.

    add_penalty : bool, default=True
        Set to True, the norm of the embedding will be penalized.
        This argument is useful for the implementation of multimodal CorrNMF.
    """
    n_embeddings_other = embeddings_other.shape[1]
    of_value = 0.0
    scalar_products = embeddings_other.T.dot(embedding)

    # aux_vec not necessarily contiguous:
    # np.dot(scalar_products, aux_vec) doesn't work
    for i in range(n_embeddings_other):
        of_value += scalar_products[i] * aux_vec[i]

    # works for alpha and beta being a scalar or vector
    of_value -= np.sum(np.exp(alpha + beta + scalar_products))

    if add_penalty:
        of_value -= np.dot(embedding, embedding) / (2 * sigma_sq)

    return -of_value


@njit
def gradient_embedding(
    embedding,
    embeddings_other,
    alpha,
    beta,
    sigma_sq,
    summand_grad,
    add_penalty=True,
):
    r"""
    The negative gradient of the objective function w.r.t. a signature or
    sample embedding in CorrNMF.

    Parameters
    ----------
    embedding : np.ndarray of shape (dim_embeddings,)
        The signature or sample embedding

    embeddings_other : np.ndarray of shape (dim_embeddings, n_samples | n_signatures)
        If 'embedding' is a signature embedding, 'embeddings_other' are
        all sample embeddings. If 'embedding' is a sample embedding,
        'embeddings_other' are all signature embeddings.

    alpha : float | np.narray of shape (n_samples,)
        If 'embedding' is a signature embedding, 'alpha' are
        all sample biases. If 'embedding' is a sample embedding,
        'alpha' is the bias of the corresponding sample.

    beta : np.array of shape (n_signatures,) | float
        If 'embedding' is a signature embedding, 'beta' is
        the corresponding signature bias. If 'embedding' is a sample embedding,
        'beta' are all signature biases.

    sigma_sq : float
        model variance

    summand_grad : np.ndarray of shape (dim_embeddings,)
        A signature/sample-independent summand of the gradient.

    add_penalty : bool, default=True
        Set to True, the norm of the embedding will be penalized.
        This argument is useful for the implementation of multimodal CorrNMF.
    """
    scalar_products = embeddings_other.T.dot(embedding)
    # works for alpha and beta being a scalar or vector
    gradient = -np.sum(
        np.exp(alpha + beta + scalar_products) * embeddings_other, axis=1
    )
    gradient += summand_grad

    if add_penalty:
        gradient -= embedding / sigma_sq

    return -gradient


@njit
def hessian_embedding(
    embedding,
    embeddings_other,
    alpha,
    beta,
    sigma_sq,
    outer_prods_embeddings_other,
    add_penalty=True,
):
    r"""
    The negative Hessian of the objective function w.r.t. a signature or
    sample embedding in CorrNMF.

    Parameters
    ----------
    embedding : np.ndarray of shape (dim_embeddings,)
        The signature or sample embedding

    embeddings_other : np.ndarray of shape (dim_embeddings, n_samples | n_signatures)
        If 'embedding' is a signature embedding, 'embeddings_other' are
        all sample embeddings. If 'embedding' is a sample embedding,
        'embeddings_other' are all signature embeddings.

    alpha : float | np.narray of shape (n_samples,)
        If 'embedding' is a signature embedding, 'alpha' are
        all sample biases. If 'embedding' is a sample embedding,
        'alpha' is the bias of the corresponding sample.

    beta : np.array of shape (n_signatures,) | float
        If 'embedding' is a signature embedding, 'beta' is
        the corresponding signature bias. If 'embedding' is a sample embedding,
        'beta' are all signature biases.

    sigma_sq : float
        model variance

    aux_vec : np.ndarray of shape (n_signatures | n_samples,)
        A row or column of
        aux[k, d] = \sum_v X_vd * p_vkd,
        where X is the data matrix and p are the auxiliary parameters of CorrNMF.
        If 'embedding' is a signature embedding, the corresponding row is provided.
        If 'embedding' is a sample embedding, the corresponding column is provided.

    add_penalty : bool, default=True
        Set to True, the norm of the embedding will be penalized.
        This argument is useful for the implementation of multimodal CorrNMF.
    """
    dim_embeddings, n_embeddings_other = embeddings_other.shape
    scalar_products = embeddings_other.T.dot(embedding)
    scalings = np.exp(alpha + beta + scalar_products)
    hessian = np.zeros((dim_embeddings, dim_embeddings))

    for m1 in range(dim_embeddings):
        for m2 in range(dim_embeddings):
            for i in range(n_embeddings_other):
                hessian[m1, m2] -= scalings[i] * outer_prods_embeddings_other[i, m1, m2]
            if add_penalty and m1 == m2:
                hessian[m1, m2] -= 1 / sigma_sq

    return -hessian



import numpy as np
import pandas as pd
from numba import njit

from ..utils import normalize_WH
from ._utils_klnmf import kl_divergence, poisson_llh, samplewise_kl_divergence, update_H
from .nmf import NMF

EPSILON = np.finfo(np.float32).eps


@njit
def volume_logdet(W: np.ndarray, delta: float) -> float:
    n_signatures = W.shape[1]
    diag = np.diag(np.full(n_signatures, delta))
    volume = np.log(np.linalg.det(W.T @ W + diag))

    return volume


@njit
def kl_divergence_penalized(
    X: np.ndarray, W: np.ndarray, H: np.ndarray, lam: float, delta: float
) -> float:
    reconstruction_error = kl_divergence(X, W, H)
    volume = volume_logdet(W, delta)
    loss = reconstruction_error + lam * volume

    return loss


@njit
def update_W_unconstrained(
    X: np.ndarray,
    W: np.ndarray,
    H: np.ndarray,
    lam: float,
    delta: float,
    n_given_signatures: int = 0,
) -> np.ndarray:
    n_signatures = W.shape[1]
    diag = np.diag(np.full(n_signatures, delta))
    Y = np.linalg.inv(W.T @ W + diag)
    Y_minus = np.maximum(0, -Y)
    Y_abs = np.abs(Y)
    WY_minus = W @ Y_minus
    WY_abs = W @ Y_abs

    rowsums_H = np.sum(H, axis=1)
    discriminant_s1 = (rowsums_H - 4 * lam * WY_minus) ** 2
    discriminant_s2 = 8 * lam * WY_abs * ((X / (W @ H)) @ H.T)
    numerator_s1 = np.sqrt(discriminant_s1 + discriminant_s2)
    numerator_s2 = -rowsums_H + 4 * lam * WY_minus
    numerator = numerator_s1 + numerator_s2
    denominator = 4 * lam * WY_abs
    W_unconstrained = W * numerator / denominator
    W_unconstrained[:, :n_given_signatures] = W[:, :n_given_signatures].copy()
    W_unconstrained[:, n_given_signatures:] = W_unconstrained[
        :, n_given_signatures:
    ].clip(EPSILON)

    return W_unconstrained


@njit
def line_search(
    X: np.ndarray,
    W: np.ndarray,
    H: np.ndarray,
    lam: float,
    delta: float,
    gamma: float,
    W_unconstrained: np.ndarray,
) -> tuple[np.ndarray, np.ndarray, float]:
    prev_of_value = kl_divergence_penalized(X, W, H, lam, delta)
    W_new, H_new = normalize_WH(W_unconstrained, H)
    W_new, H_new = W_new.clip(EPSILON), H_new.clip(EPSILON)
    of_value = kl_divergence_penalized(X, W_new, H_new, lam, delta)

    while of_value > prev_of_value and gamma > 1e-16:
        gamma *= 0.8
        W_new = (1 - gamma) * W + gamma * W_unconstrained
        W_new, H_new = normalize_WH(W_new, H)
        W_new, H_new = W_new.clip(EPSILON), H_new.clip(EPSILON)
        of_value = kl_divergence_penalized(X, W_new, H_new, lam, delta)

    gamma = min(1.0, 1.2 * gamma)

    return W_new, H_new, gamma


class MvNMF(NMF):
    """
    Min-volume non-negative matrix factorization. This algorithms is a volume-
    regularized version of NMF with the generalized Kullback-Leibler (KL)
    divergence.

    Parameters
    ----------
    n_signatures: int
        Number of signatures to decipher.

    init_method : str, default=nndsvd
        One of "custom", "flat", "hierarchical_cluster", "nndsvd",
        "nndsvda", "nndsvdar" "random" and "separableNMF". Please see the initialization
        module for further details on each method.

    lam : float, default=1.0
        Objective function volume penalty weight.

    delta : float, default=1.0
        Objective function hyperparameter.

    min_iterations : int, default=500
        Minimum number of iterations.

    max_iterations : int, default=10000
        Maximum number of iterations.

    conv_test_freq: int
        The frequency at which the algorithm is tested for convergence.
        The objective function value is only computed every 'conv_test_freq'
        many iterations, which also affects a potentially saved history of
        the objective function values.

    tol : float, default=1e-7
        Tolerance of the stopping condition.

    Reference
    ---------
    Leplat, V., Gillis, N. and Ang, A.M., 2020.
    Blind audio source separation with minimum-volume beta-divergence NMF.
    IEEE Transactions on Signal Processing, 68, pp.3400-3410.
    """

    def __init__(
        self,
        n_signatures=1,
        init_method="nndsvd",
        lam=1.0,
        delta=1.0,
        min_iterations=500,
        max_iterations=10000,
        conv_test_freq=10,
        tol=1e-7,
    ):
        super().__init__(
            n_signatures,
            init_method,
            min_iterations,
            max_iterations,
            conv_test_freq,
            tol,
        )
        self.lam = lam
        self.delta = delta
        self._gamma = None

    @property
    def reconstruction_error(self):
        return kl_divergence(self.X, self.W, self.H)

    @property
    def samplewise_reconstruction_error(self):
        return samplewise_kl_divergence(self.X, self.W, self.H)

    def objective_function(self):
        return kl_divergence_penalized(self.X, self.W, self.H, self.lam, self.delta)

    @property
    def objective(self) -> str:
        return "minimize"

    def loglikelihood(self) -> float:
        return poisson_llh(self.X, self.W, self.H)

    def _update_H(self):
        self.H = update_H(self.X, self.W, self.H)

    def _update_W_unconstrained(self):
        return update_W_unconstrained(
            self.X, self.W, self.H, self.lam, self.delta, self.n_given_signatures
        )

    def _line_search(self, W_unconstrained):
        self.W, self.H, self._gamma = line_search(
            self.X,
            self.W,
            self.H,
            self.lam,
            self.delta,
            self._gamma,
            W_unconstrained,
        )

    def _update_W(self):
        W_unconstrained = self._update_W_unconstrained()
        self._line_search(W_unconstrained)

    def fit(
        self,
        data: pd.DataFrame,
        given_signatures=None,
        init_kwargs=None,
        history=False,
        verbose=0,
    ):
        """
        Parameters
        ----------
        data : pd.DataFrame
            Input count data.

        given_signatures : pd.DataFrame, default=None
            Signatures to fix during the inference.

        init_kwargs: dict
            Any further keyword arguments to be passed to the initialization method.
            This includes, for example, a possible 'seed' keyword argument
            for all stochastic methods.

        history : bool, default=True
            If true, the objective function value of each iteration is saved.

        verbose : int, default=0
            Verbosity level.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        self._setup_data_parameters(data)
        self._initialize(given_signatures, init_kwargs)
        self._gamma = 1.0
        of_values = [self.objective_function()]
        n_iteration = 0
        converged = False

        while not converged:
            n_iteration += 1

            if verbose and n_iteration % 1000 == 0:
                print(f"iteration: {n_iteration}; objective: {of_values[-1]:.2f}")

            self._update_H()

            if self.n_given_signatures < self.n_signatures:
                self._update_W()

            if n_iteration % self.conv_test_freq == 0:
                prev_of_value = of_values[-1]
                of_values.append(self.objective_function())
                rel_change = (prev_of_value - of_values[-1]) / prev_of_value
                converged = rel_change < self.tol and n_iteration >= self.min_iterations

            converged |= n_iteration >= self.max_iterations

        if history:
            self.history["objective_function"] = of_values[1:]

        return self



"""
Salamander: a non-negative matrix factorization framework for signature analysis
================================================================================
"""
from .nmf_framework.corrnmf_det import CorrNMFDet
from .nmf_framework.klnmf import KLNMF
from .nmf_framework.multimodal_corrnmf import MultimodalCorrNMF
from .nmf_framework.mvnmf import MvNMF
from .plot import set_salamander_style

__version__ = "0.3.1"
__all__ = ["CorrNMFDet", "KLNMF", "MvNMF", "MultimodalCorrNMF"]


set_salamander_style()
